#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "../common/gui-commands.h"

#ifndef CLI_DEBUG
#define CLI_DEBUG  1
#endif

extern char *optarg;  /* argument from getopt */

void usage()
{
  printf("xsup_cli -- Command Line Interface Controller\n");
  printf("            for XSupplicant\n\n\n");
  printf("      -e <interface>:<setting>=<value> : Set the value of <setting> to <value>.\n");
  printf("      -g <interface>:<setting>   : Get the value of <setting> on interface <interface>.\n");
  printf("      -s <interface>             : Supplicant Status for Interface.\n");
  printf("      -i                         : List interfaces the daemon is using.\n");
  printf("      -h                         : Get the highest state that any interface is in.\n");
  printf("\n\n\n");
}

// Split the <interface> from the <setting>
int split_cmd(char *inval, char *get_net, char *get_arg)
{
  int split_point = 0, i;

  i=0;

  // Make sure we have two values.
  while ((i<=strlen(inval)) && (inval[i] != ':')) 
    {
      i++;
    }

  if (inval[i] != ':') return -1;  // We don't have two pieces!

  for (i=0; i<=strlen(inval); i++)
    {
      if (split_point == 0)   // We are on the network piece.
	{
	  if (inval[i] != ':')
	    {
	      get_net[i] = inval[i];
	    } else {
	      split_point = i+1;  // Start the next half beyond the ':'
	      get_net[i] = '\0';
	    }
	} else {
	  get_arg[i-split_point] = inval[i];
	}
    }
  get_arg[i+1] = '\0';

  return 0;
}

int set_status(char *command)
{
  int i;
  char *answer=NULL, *message=NULL;

  printf("Getting status for %s!\n",command);

  i = command_status(command);
  if (i!=0)
    {
      printf("Error : %s\n",get_command_error(i));
      return -1;
    }

  sleep(1);
  answer = (char *)malloc(1024);
  if (command_get_response(answer)!=0)
    {
      printf("Error getting response!\n");
      return -3;
    }

  i=0;
  while ((answer[i]!=' ') && (i<=strlen(answer))) i++;
  
  message = (char *)malloc(strlen(answer)-i);
  memcpy(message, &answer[i+1], strlen(answer)-i);

  printf("Current supplicant status is : %s\n", message);
  free(message);
  free(answer);
  message = NULL;
  answer = NULL;
  return 0;
}

int get_highest_state()
{
  int i;
  char *answer=NULL, *message=NULL;

  printf("Getting highest status!\n");

  i = command_hi_stat();
  if (i!=0)
    {
      printf("Error : %s\n",get_command_error(i));
      return -1;
    }

  sleep(1);
  answer = (char *)malloc(1024);
  if (command_get_response(answer)!=0)
    {
      printf("Error getting response!\n");
      return -3;
    }

  i=0;
  while ((answer[i]!=' ') && (i<=strlen(answer))) i++;
  
  message = (char *)malloc(strlen(answer)-i);
  memcpy(message, &answer[i+1], strlen(answer)-i);

  printf("Highest interface status is : %s\n", message);
  free(message);
  free(answer);
  message = NULL;
  answer = NULL;
  return 0;
}

int get_ints()
{
  int i;
  char *answer=NULL, *message=NULL;

  i = command_get_ints();
  if (i!=0)
    {
      printf("Error : %s\n",get_command_error(i));
      return -1;
    }

  sleep(1);
  answer = (char *)malloc(1024);
  if (command_get_response(answer)!=0)
    {
      printf("Error getting response!\n");
      return -3;
    }

  i=0;
  while ((answer[i]!=' ') && (i<=strlen(answer))) i++;
  
  message = (char *)malloc(strlen(answer)-i);
  memcpy(message, &answer[i+1], strlen(answer)-i);

  printf("Current known interfaces are : %s\n", message);
  free(message);
  free(answer);
  message = NULL;
  answer = NULL;
  return 0;
}

int get_value(char *instr)
{
  char *interface, *request, *answer;
  int tag;

  interface = (char *)malloc(strlen(instr));
  request = (char *)malloc(strlen(instr));
  
  if (split_cmd(instr, interface, request) != 0)
    {
      printf("Invalid format for argument, %s\n", instr);
      free(interface);
      interface=NULL;
      free(request);
      request=NULL;
      return -1;
    }
  if (command_get(interface, request) == 0)
    {
      answer = (char *)malloc(1024); // It should never get bigger than this!
      sleep(1);   // Give it some time to answer.
      if (command_get_response(answer) != 0)
	{
	  printf("There was an error in the answer from the daemon!\n");
	  free(answer);
	  return -3;
	}
      free(interface);
      free(request);
      interface=NULL;
      request=NULL;
      split_result(answer, &tag, &interface, &request);
      printf("Returned : %s\n",request);
    }
  return 0;
}

int set_value(char *instr)
{
  char *interface, *request, *answer;
  int tag;

  interface = (char *)malloc(strlen(instr));
  request = (char *)malloc(strlen(instr));
  
  if (split_cmd(instr, interface, request) != 0)
    {
      printf("Invalid format for argument, %s\n", instr);
      free(interface);
      interface=NULL;
      free(request);
      request=NULL;
      return -1;
    }
  if (command_set(interface, request) == 0)
    {
      answer = (char *)malloc(1024); // It should never get bigger than this!
      sleep(1);   // Give it some time to answer.
      if (command_get_response(answer) != 0)
	{
	  printf("There was an error in the answer from the daemon!\n");
	  free(answer);
	  return -3;
	}
      free(interface);
      free(request);
      interface=NULL;
      request=NULL;
      split_result(answer, &tag, &interface, &request);
      printf("Returned : %s\n",request);
    }
  return 0;
}

int main(int argc, char *argv[])
{
  int op, i;
  char *theOpts = "s:g:e:ih";

  if (argc<2)     // We need to have at least one command passed, to do anything.
    {
      usage();
      exit(-1);
    }

  i = command_connect();   // Establish connection to the daemon.
  if (i != 0)
    {
      printf("Error : %s\n", get_command_error(i));
      printf("We cannot continue!\n");
      exit(1);
    }

  while ((op = getopt(argc, argv, theOpts)) != EOF)
    {
      switch (op) {
      case 'i':  // Get a list of interfaces.
	switch (get_ints())
	  {
	  case -1:
	    printf("Bad Argument!\n");
	    break;
	  case -2:
	    printf("Send error!\n");
	    break;
	  }
	break;
      case 's':  // Get the current status of the authentication.
	switch (set_status(optarg))
	  {
	  case -1:
	    printf("Bad Argument!\n");
	    break;
	  case -2:
	    printf("Send error!\n");
	    break;
	  case 0:
	    break;
	  }
	break;
      case 'g':  // Get the value of a setting.
	switch (get_value(optarg))
	  {
	  case -1:
	    printf("Bad Argument!\n");
	    break;
	  case -2:
	    printf("Send error!\n");
	    break;
	  case 0:
	    break;
	  }
	break;
      case 'e':  // Set a value.
	switch (set_value(optarg))
	  {
	  case -1:
	    printf("Bad Argument!\n");
	    break;
	  case -2:
	    printf("Send error!\n");
	    break;
	  case 0:
	    break;
	  }
	break;
      case 'h':
	switch (get_highest_state())
	  {
	  case -1:
	    printf("Bad Argument!\n");
	    break;
	  case -2:
	    printf("Send error!\n");
	    break;
	  case 0:
	    break;
	  }
	break;
      default:
	printf("Unknown command!!!\n\n\n");
	usage();
	break;
      }
    }
  command_close();  // We are done!
  return 0;
}
