#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <fcntl.h>
#include "gui-commands.h"

#define DAEMON_CONNECTION "127.0.0.1"
#define PORT_NUM 10240

int sockfd, reqPending;
struct sockaddr_in dest;
char buffer[1024];

/* This function takes the pieces of a command, puts them together, and returns
   a string that contains a valid message to send to the supplicant. */
char *build_command(int tag, char *interface, char *message)
{
  int msgLength;
  char *to_send = NULL;

  msgLength = 3;   /* Room for a command tag, space, and NULL. */
  if (interface == NULL) return NULL;
  msgLength +=strlen(interface);

  if (message != NULL) msgLength +=strlen(interface);

  to_send = (char *)malloc(msgLength);
  bzero(to_send, msgLength);

  to_send[0] = tag;
  strncpy(&to_send[1], interface, strlen(interface));
  to_send[strlen(interface)+1] = ' ';

  if (message != NULL) strncpy(&to_send[strlen(interface)+2], message, strlen(message));

  return to_send;
}

/* This call will return the interface name that was returned to us from the
   supplicant.  It is up to the caller to decide if this is a valid interface
   or not. (But, it should be. ;) */
char *get_int_name(char *message)
{
  int i;
  char *intStr;

  i = 1;
  while ((message[i] != ' ') && (i<=strlen(message))) i++;
  
  intStr = (char *)malloc(i-1);
  memcpy(intStr, &message[1], i-1);
  intStr[i] = '\0';

  return intStr;
}

/* This call will return the piece of a message that trails the space in a
   response from the supplicant.  Some reponses may not have a message piece,
   in which case, this should return NULL.  */
char *get_message(char *instr)
{
  int i;
  char *msgStr;

  i=1;
  while ((instr[i]!=' ') && (i<=strlen(instr))) i++;

  i++;
  msgStr = (char *)malloc(strlen(instr)-i+1);

  memcpy(msgStr, &instr[i], (strlen(instr)-i+1));
  return msgStr;
}

/* A function to make it easy to get the component pieces of a response from
   the supplicant.  */
int split_result(char *instr, int *tag_value, char **interface, char **value)
{
  *tag_value = instr[0];
  *interface = get_int_name(instr);
  *value = get_message(instr);
  return 0;
}

/* Return text for an error that was passed back by one of the functions.  */
char *get_command_error(int errnum)
{
  static char *error_list[] = {
    "There was no error!",
    "Couldn't establish connection to daemon.",
    "Invalid command name!",
    "Couldn't send 'get' command!"
  };

  return error_list[errnum];
}

/* Establish a "connection" to the supplicant.  Since this is a UDP socket,
   the only thing we can be sure of is that the socket is now open.  We have
   no guarentee that there is a supplicant listening!  */
int command_connect()
{
  struct in_addr holder;
  int sockopt;

  /* If we are reconnecting, then there shouldn't be any requests pending. */
  reqPending = 0;

  /* Create a socket handler. */
  if ( (sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
      return 1;    /* Return error 1. (See get_commands_error)  */
    }

  sockopt = fcntl(sockfd, F_GETFL, 0);
  if (fcntl(sockfd, F_SETFL, sockopt | O_NONBLOCK) == -1)
    {
      return 1;
    }

  /* Initalize the socket to connect to the correct place.  */
  bzero(&dest, sizeof(dest));
  dest.sin_family = AF_INET;
  dest.sin_port = htons(PORT_NUM);
  memcpy(&holder, &dest.sin_addr.s_addr, 4);
  if ( inet_aton(DAEMON_CONNECTION, &holder) == 0)
    {
      return 1;
    }

  /* We are now done.  */
  return 0;
}

/* Close the currently open socket that we are using to talk to the supplicant.
 */
void command_close()
{
  close(sockfd);
}

/* This is an *internal* command, and shouldn't be called outside this file! */
int send_cmd(char *cmd_to_send, int length)
{
  int destlen,returnval;

  reqPending++;             /* We have a request pending. */
  destlen = sizeof(dest);

  returnval = sendto(sockfd, cmd_to_send, length, 0, (struct sockaddr *)&dest, destlen);
  if (returnval < length)
    {
      return 3;  /* Couldn't send "get" message. */
    }

  return 0;
}

/* This function will return a response from the supplicant, if there is one.
   If there isn't anything to answer with, -1 will be returned. */
int command_get_response(char *buffer)
{
  int data_size;
  char *internal_buffer;

  internal_buffer = (char *)malloc(1024);
  data_size = recvfrom(sockfd, internal_buffer, 1024, 0, NULL, NULL);

  if (data_size <= 0) 
    {
      if (reqPending >=3) 
	{
	  command_close();
	  command_connect(); /* Reconnect if we didn't get an answer. */
	} else {
	  reqPending++;
	}
      return -1;
    }

  memcpy(buffer, internal_buffer, data_size);
  free(internal_buffer);

  if ((data_size < 1024) || (buffer[data_size-1] = '+')) /* + means there is more. */
    {
      return 0;
    } 
  return 1;
}

/*  
    command_get -- Sends a request to get specific information from the 
                   daemon.  Commands that are allowed are the same values that
		   are used in the config file.
*/
  
int command_get(char *interface, char *value)
{
  int i=-1;
  char *to_send = NULL;

  to_send = build_command(GET_VALUE, interface, value);
  if (to_send != NULL)
    {
      i = send_cmd(to_send, strlen(interface)+strlen(value)+4);
      free(to_send);
    }
  return i;
}

/* command_status -- Requests the current status of the supplicant.
   Returns 0 on success, -1 on unknown error, >1 on known error.
   The value passed in should be a specific interface.  */
int command_status(char *interface)
{
  int i=-1;
  char *to_send = NULL;

  to_send = build_command(GET_STATE, interface, NULL);
  if (to_send != NULL)
    {
      i = send_cmd(to_send, strlen(interface)+3);
      free(to_send);
    }
  return i;
}

int command_hi_stat()
{
  int i=0;
  char *to_send = NULL, interface[5] = "na \0";

  to_send = (char *)malloc(strlen(&interface)+3);
  bzero(to_send, strlen(&interface)+3);

  to_send[0] = GET_HI_STATE;
  strncpy(&to_send[1], &interface, strlen(&interface));
  to_send[strlen(&interface)+1] = ' ';
  //  strncpy(to_send, stat_cmd, strlen(stat_cmd));

  i = send_cmd(to_send, strlen(&interface)+3);
  return i;
}

/* This function will ask the supplicant to give us a comma seperated list
   of all of the interfaces is it currently aware of.  */
int command_get_ints()
{
  int i=0;
  char *to_send = NULL, int_name[5] = "na \0";
  
  to_send = (char *)malloc(10);
  bzero(to_send, 10);
  
  to_send[0] = GET_INTS;
  strncpy(&to_send[1], int_name, strlen(int_name));
  
  i = send_cmd(to_send, strlen(int_name)+3);
  return i;
}

/* command_set -- Sets a value, returns 0 on success, -1 on unknown error,  */
int command_set(char *interface, char *message)
{
  int i;
  char *to_send = NULL;

  to_send = (char *)malloc(strlen(interface)+strlen(message)+3);
  bzero(to_send, strlen(interface)+strlen(message)+3);

  to_send[0] = SET_VALUE;
  strncpy(&to_send[1], interface, strlen(interface));
  to_send[strlen(interface)+1] = ' ';

  strncpy(&to_send[strlen(interface)+2], message, strlen(message));

  i = send_cmd(to_send, strlen(interface)+strlen(message)+4);
  return i;
}
