/**
 * A client-side 802.1x implementation supporting EAP/TLS
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 *
 * Copyright (C) 2002 Bryan D. Payne & Nick L. Petroni Jr.
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
 * The driver function for a Linux application layer EAPOL 
 * implementation
 * File: 1xdaemon.c
 *
 * Authors: Chris.Hessing@utah.edu, Terry.Simons@utah.edu
 *
 * $Id: 1xdaemon.c,v 1.26 2003/09/18 23:03:38 chessing Exp $
 * $Date: 2003/09/18 23:03:38 $
 * $Log: 1xdaemon.c,v $
 * Revision 1.26  2003/09/18 23:03:38  chessing
 * Memory leak fixes, inital implementation of EAP-SIM (draft 5 only).  EAP-SIM still has a lot of bugs to be fixed, and doesn't key correctly yet.  EAP-SIM also requires libpcsc, and pthread.  To enable EAP-SIM use the --enable-eap-sim flag on configure.
 *
 * Revision 1.25  2003/08/28 23:29:13  chessing
 * Added patches from mark@glines.org
 *
 * Revision 1.24  2003/07/05 06:22:28  chessing
 * Fixes to int_hooks, and userconf stuff.
 *
 * Revision 1.23  2003/07/05 05:05:21  chessing
 * Some changes to hardware inspection code, to fix some strange bugs.
 *
 * Revision 1.22  2003/07/04 04:32:41  chessing
 * Fixes for Mac OS-X (and maybe *BSD systems)..  It should build on OS-X again.
 *
 * Revision 1.21  2003/07/03 04:53:06  chessing
 * Push updates for GUIs, a few strange bug fixes.
 *
 * Revision 1.20  2003/06/29 20:16:47  chessing
 * Various code cleanups, and memory leak fixes.
 *
 * Revision 1.19  2003/06/29 05:38:48  chessing
 * Various memory leak cleanups, first hack at PEAP version 0, works with Radiator!
 *
 * Revision 1.18  2003/06/19 02:32:00  chessing
 * Additions to better support daemon mode.  Updates to the gui-commands files.
 *
 * Revision 1.17  2003/06/16 05:02:52  chessing
 * Code clean-ups, and documentation of functions.
 *
 *******************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "1xdaemon.h"
#include "logging.h"
#include "eapol.h"
#include "int_hooks.h"
#include "userconf.h"

#if MEMWATCH
#include "memwatch.h"
#endif

int msgSock    = -1;
int serverSock = -1;
struct client_sockets *startCliPtr=NULL, *curCliPtr=NULL;

int sa_length  =  0;
struct sockaddr_in caller;
const int MSGBUFMAX = 512;

// Set up the socket that we will listen to GUI interfaces from.
int serverSockInit()
{
  // For socket related calls.
  int sockErr = 0;

  // Mode, addr, and port data for our socket
  struct sockaddr_in sa;

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = INADDR_ANY;
  sa.sin_port   = htons(10240);

  // We're ready to begin authentication... set up a socket for a client frontend to connect to:
  serverSock = socket(PF_INET, SOCK_DGRAM, 0);

  // Check for errors in the socket creation
  if(serverSock == -1) {
    xlogf(DEBUG_NORMAL, DONT_FORCE, "An error occured while attempting to setup the client socket in 1xdriver.c\nError: %d\n", errno);
  } else {
    int sockOpts = fcntl(serverSock, F_GETFL, 0);
    // Set our socket to non-blocking
    sockErr = fcntl(serverSock, F_SETFL, sockOpts | O_NONBLOCK);
    if(sockErr == -1) {
      xlogf(DEBUG_NORMAL, DONT_FORCE, "An error occured while attempting to set non-blocking mode in 1xdriver.c\nError: %d\n", errno);
    } else {
      sockErr = bind(serverSock, (struct sockaddr *)&sa, sizeof(sa));
      if(sockErr == -1) {
        xlogf(DEBUG_NORMAL, DONT_FORCE, "An error occured while attempting to bind socket in 1xdriver.c\nError: %d\n", errno);
      }
    }
  }

  return serverSock;
}

void closeServerSock()
{
  close(serverSock);
}

void lower_str(char *instr)
{
  int i;

  if (instr == NULL) return;

  for (i=0; i<=strlen(instr); i++)
    {
      instr[i] = tolower(instr[i]);
    }
}

// Register a GUI client, so that we can push messages to it later.
void register_gui_client(struct sockaddr *add_gui_cli, int add_gui_cli_length)
{
  struct sockaddr_in *cur_addr, *new_addr;

  if (startCliPtr == NULL)
    {
      startCliPtr = (struct client_sockets *)malloc(sizeof(struct client_sockets));
      curCliPtr = startCliPtr;
    } else {
      curCliPtr = startCliPtr;

      // Find the tail of the list.
      while (curCliPtr->next != NULL) 
	{
	  // Make sure we don't already have this one registered.
	  cur_addr = (struct sockaddr_in *)curCliPtr->addr;
	  new_addr = (struct sockaddr_in *)add_gui_cli;
	  if (cur_addr->sin_port == new_addr->sin_port) return;
	  curCliPtr = curCliPtr->next;
	}

      curCliPtr->next = (struct client_sockets *)malloc(sizeof(struct client_sockets));
      curCliPtr = curCliPtr->next;
    }

  curCliPtr->addr = (struct sockaddr *)malloc(add_gui_cli_length);
  curCliPtr->addr_size = add_gui_cli_length;
  curCliPtr->next = NULL;

  memcpy(curCliPtr->addr, add_gui_cli, add_gui_cli_length);
}


void batch_send(char *interface, char *msg_to_send)
{
  curCliPtr = startCliPtr;

  if (startCliPtr == NULL) return;  // If there is no one to send to, don't send!
  do
    {
      sendMessage(msg_to_send[0], interface, (char *)&msg_to_send[1], curCliPtr->addr, curCliPtr->addr_size);
      curCliPtr = curCliPtr->next;
    }
  while (curCliPtr != NULL);
}

// Return the interface piece of a message that was recieved from a GUI client.
// The message format is #int value, where # is 0-255, int is the interface
// that we want information about, and value is the message piece (if any).
char *get_int_name(char *message)
{
  int i, shmid;
  char *intStr;

  // An interface name of "na" is used for requests that are not interface
  // specific.  Make sure we don't complain about it, and return NULL.
  //  if (strstr(message, "na") != NULL) return NULL;
  if ((message[1] == 'n') && (message[2] == 'a')) return NULL;

  i = 1;
  while ((message[i] != ' ') && (i<=strlen(message))) i++;

  if (i>=strlen(message))
    {
      xlogf(DEBUG_NORMAL, DONT_FORCE, "Error : get_int_name() couldn't find interface name!\n");
      return NULL;
    }
  
  intStr = (char *)malloc(i);
  bzero(intStr, i);
  memcpy(intStr, &message[1], i);
  intStr[i-1] = '\0';

  if (is_real_int(intStr))
    {
      if (get_daemon()==1)
	{
	  shmid = get_interface_shmid(intStr);
	  if (shmid < 0) return NULL;
	  user_settings_mem_attach(shmid);
	}
      return intStr;
    }
  return NULL;
}


// Return the current state for a given interface.
char *get_current_state(char *inint)
{
  char *retval=NULL;

  retval = (char *)malloc(20);  // Our response shouldn't be longer than this!
  if (retval == NULL) return NULL;

  bzero(retval, 20);

  switch (get_statemachine_state())
    {
    case LOGOFF:
      strcpy(retval, "LOGOFF");
      break;
    case DISCONNECTED:
      strcpy(retval, "DISCONNECTED");
      break;
    case CONNECTING:
      strcpy(retval, "CONNECTING");
      break;
    case ACQUIRED:
      strcpy(retval, "ACQUIRED");
      break;
    case AUTHENTICATING:
      strcpy(retval, "AUTHENTICATING");
      break;
    case HELD:
      strcpy(retval, "HELD");
      break;
    case AUTHENTICATED:
      strcpy(retval, "AUTHENTICATED");
      break;
    default:
      retval = NULL;
      break;
    }
  return retval;
}

// This function strips out the message piece of the command we recieved from
// the GUI interface.  (See get_int_name comments for info on the command 
// format.)
char *get_message(char *instr)
{
  int i;
  char *msgStr;

  if (instr == NULL) return NULL;

  i=1;
  while ((instr[i]!=' ') && (i<=strlen(instr))) i++;

  i++;
  if ((strlen(instr)-i)<=2) return NULL;

  msgStr = (char *)malloc(strlen(instr)-i+1);

  memcpy(msgStr, &instr[i], (strlen(instr)-i)+1);
  return msgStr;
}

// This function returns the current value from the userconf structure.  Before
// this is called, a call to get_int_name should be made, to make sure that
// we are looking at the right shared memory segment.
char *get_user_value(char *instr)
{
  char *retVal=NULL;

  if (instr == NULL)
    {
      xlogf(DEBUG_NORMAL, DONT_FORCE, "get_user_value() was passed in a NULL request!\n");
      return NULL;
    }

  // Convert everything to lower case, to make matching easier.
  lower_str(instr);

  // We should now have a config option name in the message variable.
  // Pass it off to be parsed, and then return the value.  It is up to the
  // GUI client to know if we are returning a numeric value, or a string.
  retVal = get_userconf_value(instr);
  return retVal;
}

// This function sets the value of a variable in the userconf structure. Before
// this is called, a call to get_int_name should be made, to make sure we
// are looking at the right shared memory segment.  On success this function
// will return the new value of the setting.  On failure, it will return NULL.
char *set_user_value(char *instr)
{
  char *retVal = NULL;

  retVal = set_userconf_value((char *)instr);

  return retVal;
}

// parse message must *ALWAYS* be called before anything else.  It will make
// sure we are in the correct shared memory location for the interface in 
// question.  parse_message will return the component parts of a command from
// the GUI interface.
void parse_message(char *inmsg, int *tagval, char **interface, char **message)
{
  *tagval = inmsg[0];
  
  *interface = get_int_name(inmsg);
  //  if ((char *)*interface == NULL) *tagval = -1;  // If it is an invalid interface, 
                                        // return an error.
  *message = get_message(inmsg);
}

// We are passed in a list of interfaces, in comma seperated format.  Take the
// list, split out each interface, and find out which one is in the "highest"
// state.  Then, return the "highest" state as a (char *).
char *get_highest_state(char *csv_intlist)
{
  char *tempstr, *piece, *state;
  int highest = 0, shmid = 0;

  tempstr = (char *)malloc(strlen(csv_intlist)+1);
  if (tempstr == NULL) return NULL;

  strcpy(tempstr, csv_intlist);

  piece = strtok(tempstr, ",");

  do
    {
      if (is_real_int(piece))
	{
	  if (get_daemon()==1)
	    {
	      shmid = get_interface_shmid(piece);
	      if (shmid < 0) return NULL;
	      user_settings_mem_attach(shmid);
	    }
	  if (get_statemachine_state()>highest) highest = get_statemachine_state();
	}
    } while ((piece = strtok(NULL, ",")) != NULL);
  
  free(tempstr);

  state = (char *)malloc(20);
  if (state == NULL) return NULL;
  bzero(state, 20);

  switch (highest)
    {
    case LOGOFF:
      strcpy(state, "LOGOFF");
      break;
    case DISCONNECTED:
      strcpy(state, "DISCONNECTED");
      break;
    case HELD:
      strcpy(state, "HELD");
      break;
    case CONNECTING:
      strcpy(state, "CONNECTING");
      break;
    case ACQUIRED:
      strcpy(state, "ACQUIRED");
      break;
    case AUTHENTICATING:
      strcpy(state, "AUTHENTICATING");
      break;
    case AUTHENTICATED:
      strcpy(state, "AUTHENTICATED");
      break;
    }
  return state;
}

void push_messages()
{
  char *tempstr, *piece, *csv_intlist, *pushmsg;
  int shmid = 0;

  csv_intlist = int_list();  // Get a CVS list of interfaces.

  tempstr = (char *)malloc(strlen(csv_intlist)+1);
  if (tempstr == NULL) return;

  strcpy(tempstr, csv_intlist);

  piece = strtok(tempstr, ",");

  do
    {
      if (is_real_int(piece))
	{
	  if (get_daemon()==1)
	    {
	      shmid = get_interface_shmid(piece);
	      if (shmid < 0) return;
	      user_settings_mem_attach(shmid);
	    }
	  pushmsg = get_push_msg();
	  if (pushmsg != NULL)  // We have a message to push.
	    {
	      batch_send(piece, pushmsg);
	      free(pushmsg);    // Free the message.
	      clear_push_msg();
	    }
	}
    } while ((piece = strtok(NULL, ",")) != NULL);
  
  free(tempstr);
  
  if (csv_intlist != NULL)
    {
      free(csv_intlist);
      csv_intlist = NULL;
    }
}

// Process any messages we have recieved.
// The format should be xxeth0 message, where xx is a #define from the .h file
// that tells us what to do.
int process_message(char *msgBuf, struct sockaddr *hostinfo, int hostinfo_length)
{
  char *interface, *message, *uservalue = NULL, *othervals;
  int tagval;
  char bogus_int[5] = "na \0";

  parse_message(msgBuf, &tagval, &interface, &message);

  switch (tagval)
    {
    case GET_VALUE:
      uservalue = get_user_value(message);
      if (uservalue != NULL)
	{
	  sendMessage(GET_RESP, interface, uservalue, hostinfo, hostinfo_length);
	}
      break;

    case SET_VALUE:
      uservalue = set_user_value(message);  
      if (uservalue != NULL)
	{
	  sendMessage(SET_RESP, interface, uservalue, hostinfo, hostinfo_length);
	}
      break;

    case GET_STATE: // This will need to be changed when multi-interface works.
      uservalue = get_current_state(interface);
      sendMessage(STATE_RESP, interface, uservalue, hostinfo, hostinfo_length);
      break;

    case CLIENT_ACK:
      xlogf(DEBUG_NORMAL, FORCE, "Client ACK isn't implemented yet!\n");
      break;

    case GET_INTS:
      uservalue = int_list();
      // We should get a bogus interface name from the client, so we need to send one back.
      sendMessage(INTS_RESP, (char *)&bogus_int, uservalue, hostinfo, hostinfo_length);
      break;
      
    case GET_HI_STATE:
      uservalue = int_list();
      othervals = get_highest_state(uservalue);
      sendMessage(GET_HI_RESP, (char *)&bogus_int, othervals, hostinfo, hostinfo_length);
      free(othervals);
      break;

    case REGISTER_GUI:
      register_gui_client(hostinfo, hostinfo_length);
      sendMessage(GUI_REGISTERED, (char *)&bogus_int, NULL, hostinfo, hostinfo_length);
      break;
      

      // We should *never* recieve these messages.
      //    case PUSH_STATE:
    case PUSH_ERR:
    case TEST_CLIENT:
      xlogf(DEBUG_NORMAL, FORCE, "Error! Supplicant recieved message from GUI client, that it should not have! (%d)\n", tagval);
      break;

    default:
      xlogf(DEBUG_NORMAL, FORCE, "Error! Unknown request from GUI client! (%d)\n", tagval);
      break;
    }
  if (interface != NULL)
    {
      free(interface);
      interface = NULL;
    }
  if (message != NULL)
    {
      free(message);
      message = NULL;
    }
  if (uservalue != NULL)
    {
      free(uservalue);
      uservalue = NULL;
    }

  // If we have a message to push, do it.
  push_messages();

  return 0;
}

// This function takes the pieces of the response to a command, builds them in
// the proper format, and returns them to the client that asked for them.
// In the future, this command may also be used to push information to the GUI
// interface.  (Things such as a status update.)
void sendMessage(char msgTag, char *interface, char *message, struct sockaddr *destination, int destlen)
{
  int msglength;
  char *message_to_send=NULL;
  char space[2] = " \0";

  msglength = 2;  // For the tag, and the space.

  if (interface != NULL) 
    {
      msglength+=strlen(interface);
    } else {
      printf("Interface is NULL!\n");
      return; // If we don't have an interface name, it isn't valid.
    }

  if (message != NULL) msglength+=strlen(message);

  message_to_send = (char *)malloc(msglength+3);
  if (message_to_send == NULL)
    {
      xlogf(DEBUG_NORMAL, FORCE, "Error allocating space to build answer to GUI!\n ");
      return;
    }

  bzero(message_to_send, msglength+1);

  message_to_send[0] = msgTag;
  strcpy(&message_to_send[1], interface);
  strcat(&message_to_send[1], (char *)&space);
  if (message != NULL) strcat(&message_to_send[1], message);

  if (sendto(serverSock, message_to_send, msglength+1, 0, destination, destlen)<msglength+1)
    {
      xlogf(DEBUG_NORMAL, DONT_FORCE, "Error : Couldn't send message to GUI client!\n");
      }
   free(message_to_send);
 message_to_send = NULL;
}

// This function checks the socket to see if there are any pending requests for
// information.  (GUI commands)  If there are, we do whatever is asked, and
// rely on them to send the correct answer.  (sendMessage should *NEVER* be
// called from here!)
int handleMessages()
{
  char msgBuffer[MSGBUFMAX];
  int read_status = -1;
  struct sockaddr fromhost;
  int fromhost_length;
  
  bzero(msgBuffer, MSGBUFMAX);
  fromhost_length = 1024;

  if((read_status = recvfrom(serverSock, msgBuffer, MSGBUFMAX, 0, (struct sockaddr *)&fromhost, &fromhost_length)) < 0)
    {
      if(errno != EWOULDBLOCK)
	{
	  xlogf(DEBUG_EVERYTHING, DONT_FORCE, "An error occured in read() in 1xdriver.c (Error: %d)\n", errno);
	}
    } else {
      process_message(msgBuffer, &fromhost, fromhost_length);
    }

  return 0;  // To keep the compiler from complaining.
}
