/**
 * A client-side 802.1x implementation supporting EAP/TLS
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 *
 * Copyright (C) 2002 Bryan D. Payne & Nick L. Petroni Jr.
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
 * The driver function for a Linux application layer EAPOL 
 * implementation
 * File: 1xdriver.c
 *
 * Authors: bdpayne@cs.umd.edu, npetroni@cs.umd.edu
 *
 * $Id: 1xdriver.c,v 1.39 2003/09/21 03:59:35 chessing Exp $
 * $Date: 2003/09/21 03:59:35 $
 * $Log: 1xdriver.c,v $
 * Revision 1.39  2003/09/21 03:59:35  chessing
 * Various code cleanups, and other updates.  Smartcard functions for EAP-SIM are more stable.
 *
 * Revision 1.38  2003/08/28 23:29:13  chessing
 * Added patches from mark@glines.org
 *
 * Revision 1.37  2003/07/05 05:05:21  chessing
 * Some changes to hardware inspection code, to fix some strange bugs.
 *
 * Revision 1.36  2003/07/03 04:53:06  chessing
 * Push updates for GUIs, a few strange bug fixes.
 *
 * Revision 1.35  2003/07/02 02:33:54  chessing
 * Moved password prompt before we send the initial responseId, and removed all password prompts from various EAP methods.
 *
 * Revision 1.34  2003/06/29 21:31:23  chessing
 * Various memory leak cleanups.
 *
 * Revision 1.33  2003/06/29 05:38:48  chessing
 * Various memory leak cleanups, first hack at PEAP version 0, works with Radiator!
 *
 * Revision 1.32  2003/06/23 20:37:01  chessing
 * Fixes to command line switches.  They now work again. 8-)
 *
 * Revision 1.31  2003/06/22 00:44:54  chessing
 * TLS related code merged in to a single file.  Should allow for easier implementation of future TLS based authentication methods.
 *
 * Revision 1.30  2003/06/19 02:32:00  chessing
 * Additions to better support daemon mode.  Updates to the gui-commands files.
 *
 * Revision 1.29  2003/06/16 05:02:52  chessing
 * Code clean-ups, and documentation of functions.
 *
 * Revision 1.28  2003/06/15 23:16:19  chessing
 * Bug fixes, and GUI Tools additions.
 *
 * Revision 1.27  2003/06/15 20:30:05  chessing
 * Children now die when parent is killed, in daemonized mode.
 *
 * Revision 1.26  2003/06/14 20:39:07  chessing
 * Added the following :
 * 	New daemon mode.  Will fork for each real interface, and start a state machine.
 * 	New GUI interface. Still needs work, but has the basics to do a few things.  Common GUI tools source can be found in the gui_tools directory.
 * 	Uses shared memory if available.  If shared memory isn't available, daemon mode is disabled.
 *
 * Revision 1.25  2003/06/13 18:13:27  chessing
 * Small fix to make -n work again.
 *
 * Revision 1.24  2003/06/12 03:33:03  chessing
 * Fixes bug where default settings would bleed through when a different network name was found, and the net network name doesn't have a value defined.
 *
 * Revision 1.23  2003/06/11 21:18:18  chessing
 * Cleaned up a lot of memory leaks, added hooks for MEMWATCH to debug.
 *
 * Revision 1.22  2003/06/08 02:57:19  chessing
 * Added logging to file support, error checking after all mallocs.
 *
 * Revision 1.21  2003/05/20 01:44:47  chessing
 * Fixed problems with the -u option.  Added patch for rekeying with Agere based wireless cards.  Cleaned up userconf.c.  And, added first hack at TTLS-CHAP. (Still doesn't work. ;)
 *
 * Revision 1.20  2003/05/14 20:45:25  chessing
 * Various small fixes, and a fix to some math in TTLS-PAP.
 *
 * Revision 1.19  2003/03/02 20:13:04  npetroni
 * increased error checking for a number of function calls
 * fixed tls to set ctx = NULL after freeing it (redundant free's segfault)
 *
 * Revision 1.18  2003/02/27 04:05:54  chessing
 * Added more error checking, fixed a few omissions that have not yet caused bugs. ;)
 *
 * Revision 1.17  2003/02/17 19:32:45  chessing
 * Added the -m (NAK) option support by default, fixed a few cosmetic bugs in the MD5 code, and fixed build process with the new 1xdaemon.c code.
 *
 * Revision 1.16  2003/02/11 00:25:21  galimorerpg
 * GCC 2.95.3 Fixes.
 *
 * Revision 1.15  2003/02/10 02:08:00  galimorerpg
 * Memory leak fix.
 *
 * Revision 1.14  2003/02/10 01:57:53  galimorerpg
 * Double free fix.
 * Code cleanups.
 *
 * Revision 1.13  2003/02/10 00:16:50  galimorerpg
 * Configuration parsing checks.  We now recover from some potential segfaults.
 *
 * Revision 1.12  2003/02/08 00:58:59  galimorerpg
 * Fixes for 1xdaemon.c (we now catch the SIGPIPE)
 *
 * Revision 1.11  2003/02/07 22:17:29  galimorerpg
 * Socket code to allow for daemonizing has been added.
 *
 * Revision 1.10  2003/02/04 20:17:38  galimorerpg
 * 02/04/03:
 *         Modifications to allow daemonization have been implemented:
 *           1xdriver.c now sets up a server socket to listen for a front end connection for clients wishing to control the authentication process.
 *           eapol.c: Fixed a bug in eapol_authenticate() that caused the main event loop in 1xdriver.c to stop functioning until the next auth cycle.
 *         ChangeLog Updates.
 *
 * Revision 1.9  2003/01/09 23:09:22  galimorerpg
 * Logging Updates.
 *
 * Revision 1.8  2003/01/09 21:01:58  galimorerpg
 * Logging Updates and Build Fixes.
 *
 * Revision 1.7  2003/01/09 20:50:15  chessing
 * Code cleanups.
 *
 * Revision 1.6  2003/01/09 20:48:32  galimorerpg
 * Logging updates.
 *
 * Revision 1.5  2003/01/09 20:17:23  galimorerpg
 * Logging Updates
 *
 * Revision 1.4  2003/01/03 22:25:35  chessing
 * Turned off debugging code, (use the configure options to turn on the debugging) and added -w option to get around some problems with the Intel iANS drivers, and some wired/wireless issues with the MacOS-X wireless code.
 *
 * Revision 1.3  2003/01/02 19:29:37  chessing
 * Update to bring the xsupplicant code current with the development work.
 *
 *
 *******************************************************************/
/***
 *** This code implements 802.1X Authentication on a supplicant
 *** and supports multiple Authentication types.  
 *** See IEEE Draft P802.1X/D11, March 27, 2001 for more details
 ***/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#ifndef NO_DAEMON
#include <sys/shm.h>
#endif

#include "eapol.h"
#include "configparse.h"
#include "userconf.h"
#include "logging.h"
#include "1xdaemon.h"
#include "int_hooks.h"

#if MEMWATCH
#include "memwatch.h"
#endif

#ifndef DEBUG_1XDRIVER
#define DEBUG_1XDRIVER 1
#endif

char pidfileName[100] =  "/var/run/";
int pidFileSet = 0;
int is_parent = 0;   // Is this the parent process?

extern char * optarg; /* argument from getopt */

// Display information on how to use this program.
void usage(char *prog) {
  char *temp;

  xlogf(DEBUG_NORMAL, FORCE_NO_LOG, "Usage: %s "
	 "[-w]"
	 "[-a authenticator_MAC_address] "
	 "[-c config file] "
	 "[-i device] "
	 "[-n network_id] "
	 "[-u user_name] "
	 "[-p password] "
	 "[-d debug_level] "
	 "[-m authentication_type] "
	 "[-l logfile name] "
	 "\n", prog);
  
  temp = (char *)return_known_eap_types();
  xlogf(DEBUG_NORMAL, FORCE_NO_LOG, "\n\nThe following EAP types are understood : %s\n", temp);
  free(temp);
  temp = NULL;
}

// Clean up whatever we can, before we quit.
void exit_cleanly(int the_signal) {
  fflush(stdout);    // Make sure everything is on the screen.
  if (is_parent) clean_ints_and_pids();
  if (pidFileSet)
    unlink(pidfileName);
  close(msgSock);        
  close(serverSock);
  eapol_shutdown();
  clean_user_conf();
  exit(0);
}

// This is where the magic happens.  We first initalize the eapol session,
// and then loop through "eapol_authenticate" to keep the state machine 
// currnet.  If we are not running as a daemon, we also handle messages
// from any GUI tools that may be speaking to us.
void do_eapol(char *device, char *netid, char *auth_addr, char *config)
{
  int ret=0;

  /**** Begin EAPOL Conversation ****/
  do {
    ret = init_eapol(device, netid, auth_addr, config);
    if ((ret == -1) && (get_daemon()!=1))
      {
	free(auth_addr);
	auth_addr = NULL;

	if (netid != NULL) 
	  {
	    free(netid);
	    netid = NULL;
	  }
	if (config != NULL)
	  {
	    free(config);
	    config = NULL;
	  }
	exit_cleanly(-1);
      } else if ((ret == -1) && (get_daemon() == 1))
	{
	  // Sleep 5 seconds, and try again.
	  sleep(5);
	}
  } while (ret != 0);

  // If we don't have enough information to start an authentication, don't.
  /*  if (get_daemon() == 1)
    {
      while (auth_ready()!=1) 
	{
	  sleep(5);
	}
	} */

  do {
    while (eapol_authenticate() == 0)  // Loop through, and try to authenticate.
      {	 
	if (get_daemon()!=1) 
	  {
	    handleMessages();  // Handle daemon front end messages.
	  }
      }

    // If we had an error, sleep for 5 seconds, and try again.
    if (get_daemon() == 1) sleep(5);
  } while (get_daemon() == 1);
}

int main(int argc, char * argv[])
{

  int ret=0, op, i;
  char * theOpts = "m:p:u:a:c:i:n:d:l:Dw";
  char *netid = NULL, *device = NULL, *config = NULL;
  char *curint = NULL;
  u_char *dstAddr = NULL;
  u_char *auth_addr = NULL;
  int m0=0,m1=0,m2=0,m3=0,m4=0,m5=0, done=0, shmid;
  char *new_username = NULL, *new_password = NULL, *new_client_type = NULL;
  char *new_auth_method = NULL;
  FILE *pidfile = NULL;
  int pid;


  config = malloc(16);
  if(config)
    {
      strcpy(config, "/etc/1x/1x.conf");
    }

  /* check for existing copies of the program */

  /**** Command Line Parsing ****/
  while ((op = getopt(argc, argv, theOpts)) != EOF) {
    switch(op) {
    case 'a':
      dstAddr = optarg;
      break;
    case 'c':
      if(config != NULL)
	{
	  free(config);
	}
      config = optarg;
      break;
    case 'd':
      debug_level = atoi(optarg);
      break;
    case 'i':
      device = optarg;
      break;
    case 'n':
      if(netid != NULL)
	{
	  free(netid);
	  netid = NULL;
	}
      netid = optarg;
      break;
    case 'D':
      set_daemon(1);
      break;
    case 'u':
      new_username = optarg;
      xlogf(DEBUG_NORMAL, DONT_FORCE, "Username override!  Using username %s\n", optarg);
      break;
    case 'p':
      new_password = optarg;
      xlogf(DEBUG_NORMAL, FORCE_NO_LOG, "Password passed in from command line!  (This isn't very secure!)\n");
      break;
    case 'm':                   
      new_auth_method = optarg;
      if (netid != NULL)
	{
	  free(netid);
	  netid = NULL;
	}
      netid = (char *)malloc(100);
      if (netid == NULL)
	{
	  xlogf(DEBUG_NORMAL, FORCE, "Couldn't allocate memory to set this connection to wired!  The connection may not work as expected.\n");
	}
      bzero(netid, 100);
      strncpy(netid, "default", 10);
      break;
    case 'w':         //Force this to be a wired connection.
      new_client_type = optarg;
      break;
    case 'l':
      init_logfile(optarg);
      break;
    default:
      usage(argv[0]);
      return -1;
    }
  }
  
  strcat(pidfileName, "xsupplicant-");
  if (device){
    strncat(pidfileName, device, 99-strlen(pidfileName));
    pidfileName[99] = '\0';
  }
  pidfile = fopen(pidfileName, "r");
  if (pidfile) {
    fscanf(pidfile, "%d", &pid);
    xlogf(DEBUG_NORMAL, FORCE, "Process already running. kill pid %d\n", pid);
    fclose(pidfile);
    
    if (config != NULL)
      {
	free(config);
	config = NULL;
      }
    exit_cleanly(-1);
  }
  
  if (dstAddr != NULL) {

    auth_addr = (u_char *)malloc(sizeof(u_char) * MY_ETHER_ADDR_LEN);

    sscanf(dstAddr, "%x:%x:%x:%x:%x:%x", &m0, &m1, &m2, &m3, &m4, &m5);
    auth_addr[0] = m0;
    auth_addr[1] = m1;
    auth_addr[2] = m2;
    auth_addr[3] = m3;
    auth_addr[4] = m4;
    auth_addr[5] = m5;

  }

  signal(SIGTERM, exit_cleanly);
  signal(SIGINT,  exit_cleanly);
  signal(SIGQUIT, exit_cleanly); 
      
  if (get_daemon()) {
    xlogf(DEBUG_NORMAL, FORCE, "Daemonizing...\n");
    if ((new_username != NULL) || (new_password != NULL) || (new_client_type != NULL) || (new_auth_method != NULL))
      {
	xlogf(DEBUG_NORMAL, FORCE, "You passed in arguments that don't make sense when running in daemon mode.  These include -u, -p, -m, and -w.  They will be ignored.\n");
      }
    pid = fork();
    
    if (pid > 0) {
      exit(0);
    }
    if (pid < 0) {
      xlogf(DEBUG_NORMAL, FORCE, "Failed to spawn daemon. exiting...");
      free(auth_addr);
      auth_addr = NULL;
      exit_cleanly(-1);
      }

    // Okay, now fork for each interface.
    init_interface_lookup();
    build_int_cache();
    cleanup_interface_lookup();

    done = 0;

    do {
      curint = get_next_interface();

      if (curint != NULL)
	{
	  device = curint;
      
	  // Get a shared memory segment for our child.
#ifndef NO_DAEMON
	  shmid = shmget(IPC_PRIVATE, sizeof(struct userconf), 0666 | IPC_CREAT);
#endif
      	  pid = fork();
	  if (pid < 0) 
	    {
	      xlogf(DEBUG_NORMAL, FORCE, "Failed to spawn controller for %s.\n",curint);
	    }
	  if (pid > 0)
	    {
	      update_int_table(device, shmid, pid);
	    }
	  initalize_user_conf(shmid);
	  if (netid == NULL)  // Which it should in daemon mode!
	    {
	      netid = (char *)malloc(100);
	      if (netid == NULL)
		{
		  xlogf(DEBUG_NORMAL, FORCE, "Cannot allocate memory for netid! This error is fatal!\n");
		  _exit(255);
		}

	      bzero(netid, 100);
	      strncpy(netid, "default", 10);
	    }
	  if(parse_all_configs(config, netid) < 0)  //Parse the config (default by default)
	    {
	      xlogf(DEBUG_NORMAL, FORCE, "A config parse error has occured.\nMake sure you are specifying your profile name (with -n) or that a profile named \"default\" exists.\n");
	      if (netid != NULL) 
		{
		  free(netid);
		  netid = NULL;
		}
	      if (config != NULL)
		{
		  free(config);
		  config = NULL;
		}
	      exit_cleanly(-1);
	    }
      
	    if (pid == 0) 
	      {
		done = 1;
		close(0);
		close(1);
		close(2);
	      }
	} else {
	  done = 1;
	}
    }
    while (done != 1);
  } else {
    init_interface_lookup();
    build_int_cache();
    cleanup_interface_lookup();

    curint = device;
    if (device == NULL)
      {
	// We don't have an interface specified, so let's use the first
	// real interface we can find.
	//	init_interface_lookup();
	curint = get_next_interface();
	device = curint;
      }

    initalize_user_conf();
    if (netid == NULL)
      {
	netid = (char *)malloc(100);
	if (netid == NULL)
	  {
	    xlogf(DEBUG_NORMAL, FORCE, "Could not allocate memory for netid.  This error is fatal!\n");
	    _exit(255);
	  }
	bzero(netid, 100);
	strncpy(netid, "default", 10);
      } else {
	eapol_set_persistant_essid(1);
      }

    parse_all_configs(config, netid);

    if (new_username != NULL) set_username(new_username, 1);
    if (new_password != NULL) set_password(new_password);
    if (new_auth_method != NULL) 
      {
	for (i=0;i<strlen(new_auth_method);i++) 
	  new_auth_method[i] = toupper(new_auth_method[i]);
	
	if (is_valid_eap(new_auth_method)!=-1)
	  {
	    set_preferred_auth(new_auth_method);
	  } else {
	    xlogf(DEBUG_NORMAL, FORCE, "Unknown EAP method %s.  It will be ignored.\n",new_auth_method);
	  }
      }
    if (new_client_type != NULL) set_client_type(new_client_type);
  }

  //  cleanup_interface_lookup();

  if (!get_daemon())
    {
      // If the server socket fails, in single interface mode, the supplicant
      // is still usable.  Print a message, and continue.
      serverSock = serverSockInit();
      if (serverSock == -1)
	{
	  xlogf(DEBUG_NORMAL, DONT_FORCE, "Server Socket Setup Failure in 1xdriver.c\n");
	  xlogf(DEBUG_NORMAL, DONT_FORCE, "GUI tools will not work!\n");
	}
      printf("Calling do_eapol, with device %s\n",device);
      do_eapol(device, netid, auth_addr, config);
      free(auth_addr);
      auth_addr = NULL;
      exit_cleanly(0);
    } else {

      if (curint != NULL)
	{
	  do_eapol(device, netid, NULL, config);  // Don't allow an auth_addr in daemon mode.
	} else {

	  serverSock = serverSockInit();  // Set up our non-blocking server socket

	  if(serverSock == -1)
	    {
	      xlogf(DEBUG_NORMAL, DONT_FORCE, "Server Socket Setup Failure in 1xdriver.c\n");
	      exit_cleanly(-1);
	    }
	  
	  is_parent = 1;   // This is the parent process.
	  while (1!=0)
	    {
	      handleMessages();
	      usleep(1000);
	    }
	}
    }

  free(auth_addr);
  auth_addr = NULL;
  exit_cleanly(ret);
  return 0;   //We never get here, but this will silence compiler warnings.
}


/** EOF **/
