#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include "peap_phase2.h"
#include "../../eap.h"
#include "../../logging.h"
#include "../../userconf.h"

#define VALID_EAP_TYPE  EAP_TYPE_MSCHAP
#define HIGHEST_PEAP_SUPPORTED   1

int inphase2_peap_version = 1;  // By default, use version 0

int set_peap_version(int new_version)
{
  if (new_version > HIGHEST_PEAP_SUPPORTED) 
    {
      inphase2_peap_version = HIGHEST_PEAP_SUPPORTED;
      return HIGHEST_PEAP_SUPPORTED;
    }

  if (inphase2_peap_version != new_version)
    {
      xlogf(DEBUG_NORMAL, DONT_FORCE, "PEAP Version changed to %d\n",new_version);
      inphase2_peap_version = new_version;
    }
  return inphase2_peap_version;
}

int get_peap_version()
{
  return inphase2_peap_version;
}

// Remove the beginning 18 bytes.
void peap_unpad_frame(u_char *in, int in_size, u_char *out, int *out_size)
{
  int i;

  *out_size = in_size - 4;

  for (i=0;i<=*out_size;i++)
    {
      out[i] = in[4+i];
    }
}

// Pad out the beginning with 18 bytes.  (Probably 0s.)
void peap_pad_frame(u_char *in, int in_size, u_char *out, int *out_size)
{
  int i;

  *out_size = in_size + 4;

  bzero(out, *out_size);
  for (i=0;i<=in_size;i++)
    {
      out[4+i] = in[i];
    }
}

void do_peap_version1(u_char *in, int in_size, u_char *out, int *out_size)
{
  char *new_frame = NULL, *username = NULL;
  int eapvalue, new_frame_size;
  uint16_t i;

  eapvalue = in[4];

  new_frame = (char *)malloc(1024);
  if (new_frame == NULL)
    {
      //      free(padded_frame);
      return;
    }
  bzero(new_frame, 1024);

  switch (eapvalue)
    {
    case EAP_REQUEST:
	// In version 1, we answer with an EAP header.
	out[0] = EAP_RESPONSE;
	out[1] = in[1];  // Use the same ID #

	username = get_phase2id();
	i = htons(strlen(username)+5);
	memcpy((char *)&out[2], (uint16_t *)&i, 2);  // The length of the username + header.
	out[4] = EAP_TYPE_ID;
	memcpy(&out[5], username, strlen(username)+1);
	*out_size = strlen(username)+5;

      break;

    case EAP_SUCCESS:
      printf("Got a phase 2 success!\n");
      break;

    case EAP_FAILURE:
      printf("Got a phase 2 failure!\n");
      break;

    case EAP_TYPE_PEAP:  // Is this a PEAP inner request?
      out[0] = EAP_RESPONSE;
      out[1] = in[1];
      out[2] = 0;
      out[3] = 6;
      out[4]=EAP_TYPE_NAK;  // NAK
      out[5]=EAP_TYPE_MSCHAP; // MS-CHAPv2
      *out_size = 6;
      break;

    case EAP_TYPE_MSCHAP: 
      eapmschap_auth_setup();  // Make sure we have everything we need.
      eapmschap_decode_packet(&in[5],in_size-5,new_frame, &new_frame_size);
      out[0] = EAP_RESPONSE;
      out[1] = in[1];

      i = ntohs(6+new_frame_size); // 6 bytes header, plus out answer
      memcpy(&out[2], (uint16_t *)&i, 2);
      out[4] = EAP_TYPE_MSCHAP;  // We have an MSCHAPv2 answer

      memcpy(&out[5], new_frame, new_frame_size);
      *out_size = new_frame_size+5;
      break;

    case PEAP_EAP_EXTENSION: // EAP Extension
      out[0] = EAP_RESPONSE;
      memcpy(&out[1], &in[1], in_size-1);
      *out_size = in_size;
      break;

    default:
      printf("Not sure how to handle this request! (We will now crash!)\n");
      break;

    }

}

void do_peap_version0(u_char *in, int in_size, u_char *out, int *out_size)
{
  char *padded_frame, *new_frame;
  int padded_size, new_frame_size, eframe = 0;

  padded_size = in_size;

  padded_frame = (char *)malloc(in_size+19);  // It is 19 bytes to pad out.
  if (padded_frame == NULL)
    {
      xlogf(DEBUG_NORMAL, FORCE, "ACK!  We can't allocate memory!\n");
      return;
    }



  if ((in[4] == 0x21) && (in[5] = 0x80))  
    {
      eframe = 1;
      memcpy(padded_frame, in, in_size);
    }

  if (eframe != 1) 
    {
      peap_pad_frame(in, in_size, padded_frame, &padded_size);
    }

  new_frame = (char *)malloc(1024);
  if (new_frame == NULL)
    {
      xlogf(DEBUG_NORMAL, FORCE, "ACK!  We can't allocate memory!\n");
      return;
    }
 
  do_peap_version1(padded_frame, padded_size, new_frame, &new_frame_size);
  if (eframe !=1) 
    {
      peap_unpad_frame(new_frame, new_frame_size, out, out_size);
    } else {
      memcpy(out, new_frame, new_frame_size);
      *out_size = new_frame_size;
    }

  eframe = 0;
}


void peap_do_phase2(u_char *in, int in_size, u_char *out, int *out_size)
{
  if (in[0] == 0x00) 
    {
      // If we are handed an ACK, then just return an ACK.
      bzero(out, 10);
      *out_size = 0;
      return;
    }

  bzero(out, 100);
  switch (get_peap_version())
    {
    case 0:
      do_peap_version0(in, in_size, out, out_size);
      break;
    case 1:
      do_peap_version1(in, in_size, out, out_size);
      break;
    }
}

