/**
 * A client-side 802.1x implementation supporting EAP/SIM
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 *
 * Copyright (C) 2003 Chris Hessing
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
* EAPOL Function implementations for supplicant
 * 
 * File: eapsim.c
 *
 * Authors: Chris.Hessing@utah.edu
 *
 * $Id: eapsim.c,v 1.10 2003/10/05 06:03:15 chessing Exp $
 * $Date: 2003/10/05 06:03:15 $
 * $Log: eapsim.c,v $
 * Revision 1.10  2003/10/05 06:03:15  chessing
 * A couple of cosmetic cleanups.
 *
 * Revision 1.9  2003/09/25 00:10:41  chessing
 * Clean up some memory leaks.
 *
 * Revision 1.8  2003/09/24 20:17:04  chessing
 * EAP-SIM code tagged to show that it was funded by Internet Foundation Austria.
 *
 * Revision 1.7  2003/09/23 04:22:54  chessing
 * Code cleanups.
 *
 * Revision 1.6  2003/09/22 05:00:20  chessing
 * EAP-SIM draft 11 implemented, and functional.
 *
 * Revision 1.5  2003/09/21 06:59:47  chessing
 * EAP-SIM (draft 5) is fully functional, including deriving keys.
 *
 * Revision 1.4  2003/09/21 06:26:59  chessing
 * Bug fixes, and changes to fips.c to make things more stable.
 *
 * Revision 1.3  2003/09/21 03:59:35  chessing
 * Various code cleanups, and other updates.  Smartcard functions for EAP-SIM are more stable.
 *
 * Revision 1.2  2003/09/19 00:26:44  chessing
 * Some code cleanup, SIM MAC checking works correctly now.
 *
 * Revision 1.1  2003/09/18 23:03:38  chessing
 * Memory leak fixes, inital implementation of EAP-SIM (draft 5 only).  EAP-SIM still has a lot of bugs to be fixed, and doesn't key correctly yet.  EAP-SIM also requires libpcsc, and pthread.  To enable EAP-SIM use the --enable-eap-sim flag on configure.
 *
 *******************************************************************/


/*******************************************************************
 *
 * The development of the EAP/SIM support was funded by Internet
 * Foundation Austria (http://www.nic.at/ipa)
 *
 *******************************************************************/


#ifdef EAP_SIM_ENABLE     // Only build this if it has been enabled.

#include <inttypes.h>
#include <openssl/hmac.h>
#include <openssl/sha.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>

#include "eapsim.h"
#include "sm_handler.h"
#include "../../configparse.h"
#include "dot1x_globals.h"
#include "eap.h"
#include "userconf.h"
#include "logging.h"
#include "auth_methods/auth_tools.h"
#include "fips.h"
#include "simd5.h"
#include "simd11.h"
#include "../tls/eapcrypt.h"

#if MEMWATCH
#include "memwatch.h"
#endif

#ifndef SIM_DEBUG
#define SIM_DEBUG  0
#endif

#define MAX_SUPPORTED_VER   1     // Maximum version we support.

char *eapsim_netid;
char *eapsim_config, *verlist, *nsres;
char noncemt[16];
char trand1[16], trand2[16], trand3[16];
char response1[65], response2[65], response3[65];
int respsize1, respsize2, respsize3, vlistlen;
char ckey1[65], ckey2[65], ckey3[65];
int ckeysize1, ckeysize2, ckeysize3;
char *keydata=NULL;
char K_int[16];                        // This is the same as the K_aut in draft 11.
uint16_t workingversion = 0;

char *do_sha1(char *tohash, int size)
{
  EVP_MD_CTX ctx;
  char *hash_ret;
  int evp_ret_len;

  hash_ret = (char *)malloc(21);  // We should get 20 bytes returned.
  if (hash_ret == NULL)
    {
      printf("There was a malloc() error in eapsim.c with hash_ret!\n");
      return NULL;
    }
 
  EVP_DigestInit(&ctx, EVP_sha1());
  EVP_DigestUpdate(&ctx, tohash, size);
  EVP_DigestFinal(&ctx, hash_ret, (int *)&evp_ret_len);

  if (evp_ret_len != 20) printf("SHA1 returned something other than what it should have!\n");

  return hash_ret;
}

int init_eapsim(char *config, char *netid)
{
#ifdef SIM_DEBUG
  xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "(EAPSIM) Initalized\n");
#endif

  eapsim_netid = netid;
  eapsim_config = config;

  return 0;
}


/*************************************************************
  eapsim_decode_packet - decode an SIM challenge, and answer it

************************************************************/

int eapsim_decode_packet(u_char *in, int in_size, u_char *out, int *out_size)
{
  int packet_offset, i, numVers, outptr=0, saved_offset = 0;
  uint16_t value16;
  char *username, *mac_calc = NULL, *framecpy = NULL;
  char mac_val[16];
  char *passwd, *hash, *sha1resp, *xseedj, *at_mac_sres;
  char K_sres[16], K_encr[16], K_recv[32], K_send[32];
  char *rethash;
  int rethashi, tlen, maxver;

  /* Initalize the smart card code, if needed. */
  if (sc_need_init() == 1) 
    {
      passwd = get_password();
      if (init_smartcard(passwd) == -1) 
	{
	  free(passwd);
	  passwd = NULL;
	  return -1;
	}

      free(passwd);
      passwd = NULL;
    }

  // The first byte will be the subtype field.  The next two are reserved.
  switch (in[0])
    {
    case SIM_START:
#if SIM_DEBUG
      printf("(EAPSIM) SIM_START!\n");
#endif
      bzero(out, 100);     // Zero out some space.
      packet_offset = 3;   // Skip the 2 reserved bytes.

      // We got a start packet, so the first part of the response is the
      // AT_NONCE_MT, so build that first.
      out[0] = SIM_START;  // Then we skip two bytes.
      out[3] = AT_NONCE_MT;
      out[4] = 5;   // Length.

      for (i=0;i<16;i++)
	{
	  noncemt[i] = rand();
	}
      outptr = 7;

      memcpy(&out[outptr], &noncemt, 16);
      outptr += 16;

      while (packet_offset <= in_size)
	{
	  // Process SIM value fields.
	  switch (in[packet_offset])
	    {
	    case AT_MAC:
	      printf("You cannot have an AT_MAC in a Start packet!\n");
	      return -1;
	      break;

	    case AT_FULLAUTH_ID_REQ:
	      if (in[packet_offset+1] != 1) printf("Length value isn't 1!\n");
	      packet_offset+=4;  // Skip the reserved, and length bytes.

	      // Build an AT_IDENTITY piece to the response.
	      out[outptr] = AT_IDENTITY; 
	      outptr++;
	      username = get_username();  // Get the ID
	      out[outptr] = (uint8_t)strlen(username);
	      outptr++;
	      i = htons(strlen(username));
	      memcpy(&out[outptr], &i, 2);
	      outptr += 2;
	      memcpy(&out[outptr], username, strlen(username));
	      outptr += strlen(username);
	      break;

	    case AT_PERMANENT_ID_REQ:
       	      printf("AT_PERMANENT_ID_REQ\n");
	      if (in[packet_offset+1] != 1) printf("Length value isn't 1!\n");
	      packet_offset+=4;  // Skip the reserved, and length bytes.

	      // Build an AT_IDENTITY piece to the response.
	      out[outptr] = AT_IDENTITY; 
	      outptr++;
	      username = get_username();  // Get the ID
	      out[outptr] = strlen(username);
	      outptr++;
	      i = strlen(username);
	      memcpy(&out[outptr], &i, 2);
	      outptr += 2;
	      memcpy(&out[outptr], username, strlen(username));
	      outptr += strlen(username);
	      break;

	    case AT_ANY_ID_REQ:
       	      printf("AT_ANY_ID_REQ\n");
	      if (in[packet_offset+1] != 1) printf("Length value isn't 1!\n");
	      packet_offset+=4;  // Skip bytes...
	      break;

	    case AT_VERSION_LIST:
	      printf("AT_VERSION_LIST\n");
	      printf("Version List Length : %d\n",in[packet_offset+1]);
	      numVers = in[packet_offset+1];
	      packet_offset+=2;

	      memcpy(&value16, &in[packet_offset], 2);
	      value16 = ntohs(value16);
	      printf("Version List Length (in bytes) : %d\n",value16);
	      packet_offset+=2;

	      verlist = (char *)malloc(value16);
	      if (verlist == NULL) return -1;
	      vlistlen = value16;
	      memcpy(verlist, &in[packet_offset], value16);
	      maxver = 0;  // Assume we are starting version 0.

	      for (i=0;i<numVers;i++)
		{
		  memcpy(&value16, &in[packet_offset], 2);
		  value16 = ntohs(value16);
		  printf("AT_VERSION_LIST Length Value : %d\n",value16);
		  if (value16 > maxver) maxver = value16;

		  packet_offset+=2;
		}
	      printf("Setting version to %d\n",maxver);
	      out[outptr] = AT_SELECTED_VERSION;
	      outptr++;
	      out[outptr] = 1;
	      outptr++;
	      out[outptr] = 0;
	      outptr++;
	      out[outptr] = maxver;  // We only do version 0.
	      outptr++;

	      workingversion = maxver;
	      break;
	      
	    case AT_SELECTED_VERSION:
	      printf("AT_SELECTED_VERSION\n");
	      if (in[packet_offset+1] != 2) printf("Length value isn't 2!\n");
	      memcpy(&value16, &in[packet_offset+2], 2);
	      value16 = ntohs(value16);
	      //printf("AT_SELECTED_VERSION Version : %d\n",value16);
	      break;
	    }
	}
      value16 = outptr;
      value16 = htons(value16);
      memcpy((char *)&out[1], &value16, 2);  // Copy the length.
      *out_size = outptr;
      break;
      
    case SIM_CHALLENGE:
#if SIM_DEBUG
      printf("(EAPSIM) SIM_CHALLENGE!\n");
#endif
      
      packet_offset = 3;
      
      out[0] = 11;  // Challenge Response
      outptr = 3;   // Where to start from.

      while (packet_offset <= in_size)
	{
	  //	  printf("packet_offset = %d\n",packet_offset);
	  // Process SIM value fields.
	  switch (in[packet_offset])
	    {
	    case AT_RAND:
	      printf("Got an AT_RAND.\n");
	      packet_offset++;
	      //printf("Length is %d\n", in[packet_offset]);
	      packet_offset+=3;  // Skip the length, and 2 reserved bytes.

	      memcpy(&trand1[0], &in[packet_offset], 16);
	      //printf("trand1 = ");
	      //print_hex((char *)&trand1[0], 16);
	      //printf("\n");
	      do_gsm(&trand1[0], &response1[0], &respsize1, &ckey1[0], &ckeysize1);
	      //printf("response1 = ");
	      //print_hex((char *)&response1[0],4);
	      //printf("\n");
	      //printf("ckey1 = ");
	      //print_hex((char *)&ckey1[0],8);
	      //printf("\n");
	      packet_offset+=16;

	      memcpy(&trand2[0], &in[packet_offset], 16);
	      //printf("trand2 = ");
	      //print_hex((char *)&trand2[0], 16);
	      //printf("\n");
	      do_gsm(&trand2[0], &response2[0], &respsize2, &ckey2[0], &ckeysize2);
	      //printf("response2 = ");
	      //print_hex((char *)&response2[0], 4);
	      //printf("\n");
	      //printf("ckey2 = ");
	      //print_hex((char *)&ckey2[0],8);
	      //printf("\n");
	      packet_offset+=16;

	      memcpy(&trand3[0], &in[packet_offset], 16);
	      //printf("trand3 = ");
	      //print_hex((char *)&trand3[0], 16);
	      //printf("\n");
	      packet_offset+=16;
	      do_gsm(&trand3[0], &response3[0], &respsize3, &ckey3[0], &ckeysize3);
	      //printf("response3 = ");
	      //print_hex((char *)&response3[0], 4);
	      //printf("\n");
	      //printf("ckey3 = ");
	      //print_hex((char *)&ckey3[0],8);
	      //printf("\n");
	      
	      if (workingversion == 0)
		{
		  hash = (char *)malloc((8*3)+16);  // 3 keys + 16 for NONCE
		  if (hash == NULL)
		    {
		      printf("Error allocating memory to build hash!\n");
		      return -1;
		    }

		  // Concatenate our keys, and NONCE.
		  bzero(hash, ((8*3)+16));
		  memcpy(&hash[0],&ckey1[0], 8);
		  memcpy(&hash[8],&ckey2[0],8);
		  memcpy(&hash[16],&ckey3[0],8);
		  memcpy(&hash[24],&noncemt[0],16);

		  //printf("Sending in to do_sha1 : ");
		  //print_hex(hash,40);
		  //printf("\n");

		  sha1resp = do_sha1(hash, 40);
		  //printf("sha1resp = ");
		  //print_hex(sha1resp,20);
		  //printf("\n");
		} else {
		  username = get_username();

		  tlen = strlen(username)+(8*3)+16+vlistlen+2;
		  printf("Size = %d\n",tlen);
		  hash = (char *)malloc(tlen);
		  if (hash == NULL) return -1;

		  nsres = (char *)malloc(4*3);
		  if (nsres == NULL) return -1;
		  bzero(nsres, 12);
		  memcpy(&nsres[0], &response1[0],4);
		  memcpy(&nsres[4], &response2[0],4);
		  memcpy(&nsres[8], &response3[0],4);

		  bzero(hash, tlen);
		  memcpy(&hash[0], &username[0], strlen(username));
		  memcpy(&hash[strlen(username)], &ckey1[0], 8);
		  memcpy(&hash[strlen(username)+8], &ckey2[0], 8);
		  memcpy(&hash[strlen(username)+16], &ckey3[0], 8);
		  memcpy(&hash[strlen(username)+24], &noncemt[0], 16);
		  memcpy(&hash[strlen(username)+24+16], verlist, vlistlen);

		  hash[strlen(username)+24+16+vlistlen+1] = workingversion;

		  printf("vlistlen = %d\n",vlistlen);

		  printf("Dump = ");
		  for (i=0;i<tlen;i++)
		    {
		      printf("%02X ",(uint8_t)hash[i]);
		    }
		  printf("\n");
		  sha1resp = do_sha1(hash, (strlen(username)+24+16+vlistlen+2));
		}
					
	      free(hash);
	      hash = NULL;

	      xseedj = (char *)malloc(20);
	      if (xseedj == NULL)
		{
		  printf("Couldn't malloc xseedj!\n");
		  return -1;
		}
	      bzero(xseedj, 20);

	      // Do the FIPS PRNG
	      at_mac_sres = do_fips186_2(sha1resp, 20, xseedj, 20, 10);
	      
	      free(xseedj);
	      xseedj = NULL;

	      //printf("at_mac_sres = ");
	      //print_hex(at_mac_sres,64);
	      //printf("\n");

	      //       	      memcpy(&K_sres[0], &at_mac_sres[0], 16);
	      if (workingversion == 0)
		{
		  for (i=0;i<16;i++) K_sres[i] = at_mac_sres[i];
     
		  memcpy(&K_encr[0], &at_mac_sres[16], 16);
		  //	      memcpy(&K_int[0], &at_mac_sres[32], 16);
		  for (i=32;i<(32+16);i++) K_int[i-32] = at_mac_sres[i];

		  //printf("K_sres = ");
		  //print_hex(&K_sres[0],16);
		  //printf("\n");
		  //printf("K_encr = %16X\n", &K_encr[0]);
		  //printf("K_int = %16X\n", &K_int[0]);

		  bzero(&K_recv[0],32);
		  bzero(&K_send[0],32);
		  for (i=48;i<(48+20);i++) K_recv[i-48] = at_mac_sres[i];
		  /*printf("K_recv = ");
		    print_hex(&K_recv[0],20);
		    printf("\n");*/
		  for (i=64;i<(68+20);i++) K_send[i-68] = at_mac_sres[i];
		  /*printf("K_send = ");
		    print_hex(&K_send[0],20);
		    printf("\n");*/
		} else {  // Version 1

		  // K_int is the same as K_aut in Version 1.
		  for (i=0;i<16;i++) K_int[i] = at_mac_sres[i+16];
		  
		  for (i=0;i<32;i++) K_recv[i] = at_mac_sres[i+32];
		  for (i=0;i<32;i++) K_send[i] = at_mac_sres[i+64];
		}
	  /*	      if (keydata != NULL)
		{
		  free(keydata);
		  keydata = NULL;
		  }*/

	      keydata = (char *)malloc(64);
	      if (keydata == NULL) return -1;

	      bzero(keydata, 64);
	      
	      memcpy(keydata, &K_recv[0],32);
      	      memcpy(&keydata[32], &K_send[0],32);

	      if (workingversion == 0)
		{
		  hash = (char *)malloc((4*3)+16);  // 3 keys + 16 for NONCE
		  if (hash == NULL)
		    {
		      printf("Error allocating memory to build hash!\n");
		      return -1;
		    }

		  // Concatenate our keys, and NONCE.
		  bzero(hash, ((4*3)+1));
		  memcpy(&hash[0],&response1[0],4);
		  memcpy(&hash[4],&response2[0],4);
		  memcpy(&hash[8],&response3[0],4);
		  hash[12] = 11;
	      
		  rethash = (char *)malloc(100);
		  if (rethash == NULL) return -1;
		  
		  HMAC(EVP_sha1(), &K_sres[0], 16, &hash[0], 13, rethash, &rethashi);
		  
		  free(hash);
		  hash = NULL;
	      
		  /*printf("Final return value : ");
		    for (i=0;i<rethashi;i++) 
		    {
		    printf("%02X ",(uint8_t)rethash[i]);
		    }
		    printf("\n");*/

		  out[outptr]=AT_MAC_SRES;
		  outptr++;

		  out[outptr]=5;
		  outptr++;
		  
		  value16 = 0;
		  memcpy(&out[outptr], &value16, 2);
		  outptr+=2;
	      
		  for (i=0;i<rethashi;i++) out[outptr+i]=rethash[i];
		  outptr+=rethashi;

		  free(rethash);
		  rethash = NULL;
		}

	      free(sha1resp);
	      sha1resp = NULL;

	      //	      printf("Offset : %d\n",packet_offset);
	      break;

	    case AT_IV:
	      printf("Got an IV (Not Supported)\n");
	      break;
	      
	    case AT_ENCR_DATA:
	      printf("Got an AT_ENCR_DATA  (Not Supported)\n");
	      break;

	    case AT_MAC:
#if SIM_DEBUG
	      printf("Got an AT_MAC\n");
#endif

	      saved_offset = packet_offset;

	      memcpy((char *)&mac_val[0], (char *)&in[packet_offset+4], 16);
	      packet_offset+=20;  // The first 16 bytes of the SHA-128 hash.
	      mac_calc = (char *)malloc(20);
	      if (mac_calc == NULL) return -1;

	      if (workingversion == 0)
		{
		  if (do_v0_at_mac(&K_int[0], in, in_size, saved_offset, mac_calc) == -1)
		    {
		      printf("Error calculating AT_MAC for Version 0!\n");
		      return -1;
		    }
		} else {
		  printf("Calculating v1 MAC.\n");
		  if (do_v1_at_mac(&K_int[0], in, in_size, saved_offset, 
				   &noncemt[0], verlist, vlistlen, 
				   workingversion, mac_calc) == -1)
		    {
		      printf("Error calculating AT_MAC for Version 1!\n");
		      return -1;
		    }
		}

	      if (memcmp(mac_calc, &mac_val[0], 16)!=0)
		{
		  printf("ERROR : AT_MAC failed MAC check!\n");
		  return -1;
		}

	      free(framecpy);
	      framecpy = NULL;

	      free(mac_calc);
	      mac_calc = NULL;

	      break;
	    }
	}

      if (workingversion == 1)   // Handle the version 1 response.
	{
	  printf("Building answer MAC.\n");
	  // First, construct the EAP packet to be used in a hash.
	  framecpy = (char *)malloc(outptr+8+20+(8*3));
	  if (framecpy == NULL) return -1;
	  
	  bzero(framecpy, (outptr+5+20+(4*3)));

	  framecpy[0] = 2;  // EAP_RESPONSE
	  framecpy[1] = get_receivedId();
	  value16 = (outptr+5+20);
	  value16 = htons(value16);   // Convert it to network byte order.
	  memcpy(&framecpy[2], &value16, 2);
	  framecpy[4] = 18;  // EAP_SIM
	  memcpy(&framecpy[5], &out[0], outptr);
	  //	  memcpy(&framecpy[5+outptr], nsres, 24);

	  framecpy[5+outptr] = AT_MAC;
	  framecpy[5+outptr+1] = 5;
	  memcpy(&framecpy[5+outptr+20], nsres,(4*3));
	  mac_calc = (char *)malloc(20);
	  if (mac_calc == NULL) return -1;

	  HMAC(EVP_sha1(), &K_int[0], 16, framecpy, (outptr+5+20+12), mac_calc, &i);
	  printf("frame = ");
	  for (i=0;i<(outptr+5+20+12);i++)
	    {
	      printf("%02X ",(uint8_t)framecpy[i]);
	    }
	  printf("\n");

	  printf("mac_calc = ");
	  for (i=0;i<20;i++)
	    {
	      printf("%02X ",(uint8_t)mac_calc[i]);
	    }
	  printf("\n");

	  memcpy(&out[outptr], &framecpy[5+outptr], 20);
	  for (i=0;i<16;i++) out[outptr+4+i] = mac_calc[i];

	  outptr+=20;
	  printf("outptr = %d\n",outptr);
	}

      value16 = outptr;
      value16 = htons(value16);
      memcpy((char *)&out[1], &value16, 2);  // Copy the length.
      *out_size = outptr;      
      break;

    case SIM_NOTIFICATION:
      printf("(EAPSIM) SIM_NOTIFICATION!  (Unsupported!)\n");
      break;

    case SIM_REAUTHENTICATION:
      printf("(EAPSIM) SIM_REAUTHENTICATION!   (Unsupported)\n");
      break;

    default:
      printf("(EAPSIM) Unknown SubType value! (%d)\n", in[0]);
      break;
    }

  out[2] = 0;  // We need to handle this better.
  return 0;  // Everything is okay.
}

/******************************************************************
 * Get the password for the <nasid:id> found in the config file.
 ******************************************************************/
int eapsim_auth_setup()
{
  return 0;
}

// Clean up any memory we have allocated, and destroy anything we need to.
int eapsim_shutdown()
{
  // Don't free one_x_globals as it is a copy of a pointer to memory, and
  // the shutdown_eap function will free it.

#ifdef SIM_DEBUG
  xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "(EAPSIM) Cleaning up.\n");
#endif

  if (sc_need_init() == 0)  // We did an init, so clean up.
    {
      if (close_smartcard() == -1) return -1;
    }

  return 0;
}

int sim_gen_keyblock()
{
  eapcrypt_set_keyblock(keydata, 64);
  return 0;
}

#endif

