/**
 * A client-side 802.1x implementation supporting EAP/SIM
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 *
 * Copyright (C) 2003 Chris Hessing
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
* EAPOL Function implementations for supplicant
 * 
 * File: fips.c
 *
 * Authors: Chris.Hessing@utah.edu
 *
 * $Id: fips.c,v 1.5 2003/09/24 20:17:04 chessing Exp $
 * $Date: 2003/09/24 20:17:04 $
 * $Log: fips.c,v $
 * Revision 1.5  2003/09/24 20:17:04  chessing
 * EAP-SIM code tagged to show that it was funded by Internet Foundation Austria.
 *
 * Revision 1.4  2003/09/23 04:22:54  chessing
 * Code cleanups.
 *
 * Revision 1.3  2003/09/21 06:26:59  chessing
 * Bug fixes, and changes to fips.c to make things more stable.
 *
 * Revision 1.2  2003/09/21 03:59:35  chessing
 * Various code cleanups, and other updates.  Smartcard functions for EAP-SIM are more stable.
 *
 * Revision 1.1  2003/09/18 23:03:38  chessing
 * Memory leak fixes, inital implementation of EAP-SIM (draft 5 only).  EAP-SIM still has a lot of bugs to be fixed, and doesn't key correctly yet.  EAP-SIM also requires libpcsc, and pthread.  To enable EAP-SIM use the --enable-eap-sim flag on configure.
 *
 *******************************************************************/

/*******************************************************************
 *
 * The development of the EAP/SIM support was funded by Internet
 * Foundation Austria (http://www.nic.at/ipa)
 *
 *******************************************************************/


/* FIPS 186-2 Implementation   */

#include <inttypes.h>
#include <strings.h>
#include <string.h>
#include <openssl/hmac.h>
#include <openssl/sha.h>
#include "sha_transform.h"


void do_add(char *inval1, char *inval2, char *result, int numlen)
{
  unsigned int toadd1, toadd2, resultvar, carry;
  int i;

  resultvar = 0;
  carry = 0;

  for (i=19;i>=0;i--)
    {
      toadd1 = 0;
      toadd2 = 0;

      toadd1 = (uint8_t)inval1[i];
      toadd2 = (uint8_t)inval2[i];

      resultvar = toadd1+toadd2+carry;
      result[i] = (resultvar & 0x00ff);
      carry = ((resultvar & 0xff00) >> 8);
    }
}

/* This is taken from the pseduocode from the FIPS 186-2 publication */
/* Page 18, Appendix 3.1  (And with modifications from pages 74-75) */

// xkeylen is the number of bytes from XKEY

// The x mod 2^(keybitlen), make sure we only return a value of (keybitlen)

char *do_fips186_2(char *XKEY, int xkeylen, char *XSEEDj, int xseedjlen, int numvals)
{
  int j,m;
  char *xval, *w1, *w2, *one, *x;  // Only return 100 values, max.

  // Do this, just to make the code match the pseudocode better.
  m = numvals;       // The number of values we want back.

  if ((xkeylen*8) != 160)
    {
      printf("Invalid XKEY length, for G(t,c) defined by SHA1!  (XKEY must be 160 when XKEY is expressed in bit length!)\n");
      return NULL;
    }

  x = (char *)malloc((20*m)*2);        // Allocate our byte array.
  xval = (char *)malloc(xkeylen);
  one = (char *)malloc(20);
  bzero(one,20);
  one[19] = 1;

  for (j=0;j<=(m-1);j++)
    {
      //We are supposed to loop through this twice, so, this is that same
      // thing, only unrolled.

      memcpy(xval, XKEY, xkeylen);
      w1 = sha_transform(xval);
	  
      // Add 1.
      do_add(XKEY, one, XKEY, xkeylen);
      do_add(XKEY, w1, XKEY, xkeylen);

      memcpy(xval, XKEY, xkeylen);
      w2 = sha_transform(xval);

      do_add(XKEY, one, XKEY, xkeylen);
      do_add(XKEY, w2, XKEY, xkeylen);
	
      memcpy(&x[j*40], w1, xkeylen);
      memcpy(&x[(j*40)+20], w2, xkeylen);
    }

  free(one);
  one = NULL;

  free(xval);
  xval = NULL;

  return x;
}

