/**
 * A client-side 802.1x implementation supporting EAP/TLS
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 * 
 * Copyright (C) 2002 Bryan D. Payne & Nick L. Petroni Jr.
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
 * EAPTLS Function implementations
 * 
 * File: eaptls.c
 *
 * Authors: bdpayne@cs.umd.edu, npetroni@cs.umd.edu
 *
 * $Id: eaptls.c,v 1.26 2003/07/02 02:33:55 chessing Exp $
 * $Date: 2003/07/02 02:33:55 $
 * $Log: eaptls.c,v $
 * Revision 1.26  2003/07/02 02:33:55  chessing
 * Moved password prompt before we send the initial responseId, and removed all password prompts from various EAP methods.
 *
 * Revision 1.25  2003/06/29 21:31:24  chessing
 * Various memory leak cleanups.
 *
 * Revision 1.24  2003/06/22 00:44:54  chessing
 * TLS related code merged in to a single file.  Should allow for easier implementation of future TLS based authentication methods.
 *
 * Revision 1.23  2003/06/11 21:19:51  chessing
 * Cleaned up memory leaks, added hooks for MEMWATCH.
 *
 * Revision 1.22  2003/06/10 22:38:53  galimorerpg
 * free() cleanups
 *
 * Revision 1.21  2003/06/08 03:34:23  chessing
 * Added error checking on TLS operations.  Should now provide information when certificates are incorrect.
 *
 * Revision 1.20  2003/06/08 02:58:16  chessing
 * Added logging to file, error checking after all mallocs.
 *
 * Revision 1.19  2003/06/06 06:07:34  chessing
 * Applied patches submitted by Brian Fundakowski Feldman from Fujitsu
 *
 * Revision 1.18  2003/05/13 19:03:30  chessing
 * Added patches from Jouni Malinen, and Patrick Sereno
 *
 * Revision 1.17  2003/04/14 00:22:05  chessing
 * Various small fixes, added new configuration options to turn off some things, and to allow static linking of OpenSSL
 *
 * Revision 1.16  2003/03/28 22:54:07  chessing
 * Patch to fix endianess issues between PPC and x86, with TTLS-PAP.
 *
 * Revision 1.15  2003/03/25 23:59:28  chessing
 * Added support for authentication method specific keyblock generation, and added dynamic WEP support for TTLS.
 *
 * Revision 1.14  2003/03/20 20:14:46  chessing
 * Many patches from Xerox PARC.  They include fixes to show-stoppers, better roaming support, CRLs and chain verification, patch to allow spaces in ESSIDs, some state machine timing issues, and prism2 card fixes.
 *
 * Revision 1.13  2003/03/11 19:30:56  chessing
 * Various code cleanups
 *
 * Revision 1.12  2003/03/11 18:06:32  chessing
 * TLS should now work with IAS and Radiator
 *
 * Revision 1.11  2003/03/10 23:35:45  chessing
 * Patches to try to get TLS working with RADIUS servers other than FreeRADIUS.
 *
 * Revision 1.10  2003/01/24 20:48:23  chessing
 * Cleaned up framing code, TLS no longer sends a malformed frame.
 *
 * Revision 1.9  2003/01/23 22:45:01  chessing
 * Added some global functions to handle packet fragments, and more TLS cleanups.
 *
 * Revision 1.8  2003/01/14 23:52:07  chessing
 * More work on the TLS code.  It should be mostly stable now.  There is a problem if get_pass("") in eaptls_auth_challenge is called twice.
 *
 * Revision 1.7  2003/01/14 19:12:50  chessing
 * TLS code now uses OpenSSL!  Cleaned out some of the no longer needed stuff from the TLS code.  Still needs more work on error checking.
 *
 * Revision 1.6  2003/01/10 22:09:08  galimorerpg
 * SSL_DEBUG define fix for gcc 3.2
 *
 * Revision 1.5  2003/01/09 20:17:29  galimorerpg
 * Logging Updates
 *
 * Revision 1.4  2003/01/09 18:38:01  galimorerpg
 * Makefile fixes
 *
 * Revision 1.3  2003/01/06 22:31:08  chessing
 * Debugging code cleanups....
 *
 * Revision 1.2  2003/01/03 22:25:36  chessing
 * Turned off debugging code, (use the configure options to turn on the debugging) and added -w option to get around some problems with the Intel iANS drivers, and some wired/wireless issues with the MacOS-X wireless code.
 *
 * Revision 1.1  2003/01/02 19:35:47  chessing
 * Add some files that were missed in the last import..
 *
 *
 *******************************************************************/
/***
 *** Specifics for EAPTLS can be found in 
 *** IETF RFC 2716
 ***/

#ifndef SSL_DEBUG
#define SSL_DEBUG 0  /* turn debug info on  deprecated- Use --with-tls-debug*/
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <openssl/err.h>

#include "eaptls.h"
#include "eapcrypt.h"
#include "tls_funcs.h"
#include "../../userconf.h"
#include "logging.h"
#include "auth_methods/auth_tools.h"

#if MEMWATCH
#include "memwatch.h"
#endif

/** GLOBAL VARS **/
char *eaptls_netid;
char *eaptls_config;


int root_cert_loaded = 0;
int user_cert_loaded = 0;

/** FUNCTION DEFINITIONS **/
 
/**
 * Initialization function for EAPTLS.  This initializes the
 * data needed for the protocol and initializes variables needed to
 * start the eaptls handshake.
 * CHANGED by npetroni to only do those things which happen *once*
 * at TLS startup. Added function eaptls_reset() to prepare TLS
 * for a new handshake
 *
 * (IN)  file name to the config file
 * (OUT) success = 0, failure = -1
 */

int init_eaptls (char *config, char *netid)
{
  eaptls_netid = netid;
  eaptls_config = config;

#if SSL_DEBUG
  xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "(EAPTLS) Changed state to 0\n");
#endif

  return tls_funcs_init();
}

// Shutdown and cleanup anything we need to.
int eaptls_shutdown()
{
  // Don't free one_x_globals here.  That will be handled by shutdown_eap.

#if SSL_DEBUG
  xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "(EAPTLS) Cleaning up.\n");
#endif

  return tls_funcs_shutdown();
}


/**
 * Decodes a packet and creates the data needed for a new packet.
 * This function is called by the eap layer with new packets as
 * they are received.  The input buffer points to the beginning
 * of the EAPTLS portion of the packet (the flags).  The reply
 * packet should be placed in the output buffer.
 *
 * The memory for the output buffer is created here, but will be 
 * freed by the eap layer when it is done using it.
 *
 * (IN)  input and output packet buffers, with lengths
 * (OUT) success = 0, failure = -1
 */
int eaptls_decode_packet (u_char *in, int in_size, u_char *out, int *out_size)
{
  return tls_funcs_decode_packet(in, in_size, out, out_size, NULL);
}

int eaptls_auth_setup()
{
  char *client_cert = NULL;
  char *client_key = NULL;
  char *temp_password = NULL;
  char *temp_username = NULL;
  char *root_cert = NULL;
  char *root_dir = NULL;
  char *crl_dir = NULL;

  temp_username = get_username();

  if (root_cert_loaded == 0)
    {

      // load CA cert.
      root_cert = get_root_cert();   //Get the filename/path for the root cert.
      root_dir = get_root_dir();     //Get a directory for our trusted roots (both for servers
	  								 // and to build our own chain
	  crl_dir = get_crl_dir();		 // where we might want to keep crls

      if (eapcrypt_tls_load_root_certs(root_cert, root_dir, crl_dir) < 0)
	{
	  if (root_cert != NULL) {free(root_cert); root_cert = NULL;}
	  if (root_dir != NULL) {free(root_dir); root_dir = NULL;}
	  if (crl_dir != NULL) {free(crl_dir); crl_dir = NULL;}
	  return -1;
	}

      if (root_cert != NULL) {free(root_cert); root_cert = NULL;}
      if (root_dir != NULL) {free(root_dir); root_dir = NULL; }
      if (crl_dir != NULL) {free(crl_dir); crl_dir = NULL;}
      root_cert_loaded = 1;  // We now have the root cert loaded.
    }

  if (user_cert_loaded == 0)
    {
      // First, get the password for the private key.
      temp_password = get_password();  //See if we have one in the file.

      if (temp_password == NULL)    // This should be impossible at this point!
        {
          if (temp_username != NULL) 
	    {
	      free(temp_username);
	      temp_username = NULL;
	    }
	  if (root_cert != NULL) {free(root_cert); root_cert = NULL;}
	  if (root_dir != NULL) {free(root_dir); root_dir = NULL;}
	  if (crl_dir != NULL) {free(crl_dir); crl_dir = NULL;}
          return -1;
        }

      // Clean up after ourselves..

      if (temp_username != NULL) 
	{
	  free(temp_username);
	  temp_username = NULL;
	}

      client_cert = get_client_cert();
      client_key = get_key_file();

      if (eapcrypt_tls_load_user_cert(client_cert, client_key, temp_password) < 0)
	{
	  xlogf(DEBUG_NORMAL, FORCE, "Couldn't load certificate! (May be an incorrect password!)\n");
	  if (client_cert)
	    {
	    free(client_cert);
	    client_cert = NULL;
	    }
	  if (client_key) 
	    {
	      free(client_key);
	      client_key = NULL;
	    }
	  return -1;
	}

      if (client_cert != NULL)
	{
	  free(client_cert);
	  client_cert = NULL;
	}

      if (client_key != NULL)
	{
	  free(client_key);
	  client_key = NULL;
	}

      set_password(temp_password);
      user_cert_loaded = 1;
      return 0;
    }
  
  if (temp_username != NULL)
    {
      free(temp_username);
      temp_username = NULL;
    }

  if (temp_password != NULL)
    {
      free(temp_password);
      temp_password = NULL;
    }

  return 0;
}

int tls_gen_keyblock()
{
  u_char *gen_keyblock;

  gen_keyblock = eapcrypt_gen_keyblock(EAPCRYPT_SESSION_KEY_CONST,
				       EAPCRYPT_SESSION_KEY_CONST_SIZE);
  if (!gen_keyblock)
    return -1;

  eapcrypt_set_keyblock(gen_keyblock, EAPCRYPT_SESSION_KEY_SIZE);
  if(gen_keyblock != NULL)
    {
      free(gen_keyblock);
      gen_keyblock = NULL;
    }

  return 0;
}

/*** EOF ***/
