/**
 * A client-side 802.1x implementation supporting EAP/TLS
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 * 
 * Copyright (C) 2002 Bryan D. Payne & Nick L. Petroni Jr.
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
 * EAPTLS Function implementations
 * 
 * File: tls_funcs.c
 *
 * Authors: Chris.Hessing@utah.edu
 *
 * $Id: tls_funcs.c,v 1.5 2003/10/20 01:09:19 chessing Exp $
 * $Date: 2003/10/20 01:09:19 $
 * $Log: tls_funcs.c,v $
 * Revision 1.5  2003/10/20 01:09:19  chessing
 *
 * Fixed segfault with PEAP.
 *
 * Revision 1.4  2003/07/02 03:40:45  chessing
 * We now check the essid each time through the authentication loop.  This should allow seamless roaming between networks with different essids.
 *
 * Revision 1.3  2003/06/29 20:16:47  chessing
 * Various code cleanups, and memory leak fixes.
 *
 * Revision 1.2  2003/06/29 05:38:48  chessing
 * Various memory leak cleanups, first hack at PEAP version 0, works with Radiator!
 *
 * Revision 1.1  2003/06/22 00:44:54  chessing
 * TLS related code merged in to a single file.  Should allow for easier implementation of future TLS based authentication methods.
 *
 *
 *******************************************************************/
#include <openssl/err.h>
#include <string.h>
#include <netinet/in.h>
#include "eaptls.h"
#include "eapcrypt.h"
#include "logging.h"
#include "tls_funcs.h"
#include "../auth_tools.h"

#if MEMWATCH
#include "memwatch.h"
#endif

int tls_funcs_phase = 0, tls_funcs_cert_size = 0;


int tls_funcs_init()
{
  return eapcrypt_tls_init();
}

int tls_funcs_shutdown()
{
  return eapcrypt_tls_shutdown();
}

int tls_funcs_decode_packet(u_char *in, int in_size, u_char *out, int *out_size, phase2_call phase2_func)
{
  int rtnVal = 0;
  u_char *p = NULL;
  uint8_t *tptr = NULL;
  uint32_t len_long;
  int tcnt;
#if TTLS_DEBUG
  int i;
#endif
  unsigned long err;
  long temp_size;
  char *temp1=NULL, *temp2=NULL;
  int temp1_size, temp2_size;

  *out_size = 0;

  err = ERR_get_error();
  if (err != 0)
    {
      xlogf(DEBUG_NORMAL, FORCE, "SSL Error -- %s\n", ERR_error_string(err, temp1));
    }

  // since we are acting as the supplicant, we can safely assume
  // that the request bit is set on all packets that we receive
  
  // see what type of packet this is
  switch(in[0]){
    case (EAPTLS_START):
#if TTLS_DEBUG
      xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "(EAPTTLS) Recieved eap-ttls start packet\n");
#endif //TTLS_DEBUG
      /* prepare to start a new handshake */
      destroy_data_frags();
      eapcrypt_tls_reset();  //This will reset the TLS.
#if TTLS_DEBUG
      xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "Changed state to INIT\n");
#endif

      rtnVal = eapcrypt_tls_parse_data(NULL, 0);

      if (rtnVal < 0)
	{
	  xlogf(DEBUG_NORMAL, FORCE, "(EAPTTLS) Parse data failed!\n");
	  return rtnVal;
	}

      rtnVal = eapcrypt_tls_return_data(out, out_size);

      tls_funcs_phase = 1;  // We are in phase 1.

      break;
//---------------------------------------------------------------------------
    case (EAPTLS_LENGTH_MORE):
    case (EAPTLS_MORE_FRAGS):
    case (EAPTLS_LENGTH_INCL):
      /* This is a fragment of data we care about.  So, stash it away.  If
         it is the final fragment, start returning data, instead of ACKs. */

      p = in + 1;

      if ((in[0] == EAPTLS_LENGTH_MORE) || (in[0] == EAPTLS_LENGTH_INCL))
	{
	  memcpy(&len_long, &in[1], 4);
	  temp_size = ntohl(len_long);
	  tls_funcs_cert_size = temp_size;
	  p=&in[5];
	  in_size-=1;
	} else {
	  in_size+=3;
	}

      /* sanity check on the given packet length */
      if (in_size == 0) {
	xlogf(DEBUG_NORMAL, FORCE, "(EAPTTLS) Packet has 0 length! (We should *NEVER* get here!\n");
	rtnVal = eapcrypt_tls_return_data(out, out_size);
	return rtnVal;
      }

      if (save_data_fragment(p, in_size) != 0)  
	{
	  xlogf(DEBUG_NORMAL, FORCE, "(EAPTTLS) Couldn't store packet fragment!\n");
	  return -1;
	} else {
	  xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "(EAPTTLS) Saved packet fragment.\n");
	}

#if TTLS_DEBUG
      xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "(EAPTTLS) Saved packet fragment\n");
#endif

      if (in[0] != EAPTLS_LENGTH_INCL)
	{
	  // we should also return an ack when we are done
	  rtnVal = tls_funcs_build_ack(out, out_size);
	} else {
	  rtnVal = eapcrypt_tls_parse_data((char *)get_data_fragment(), get_data_frag_size());
	  destroy_data_frags();
	  rtnVal = eapcrypt_tls_return_data(out, out_size);
	}

      if ((eapcrypt_state() == 0x0003) && (phase2_func != NULL))
	{
	  // This implementation of phase 2 will only work with PAP/CHAP! Other
	  // phase 2 authentication methods may require phase 2 hooks in
	  // other locations.

	  temp1 = (char *)malloc(1024);
	  if (temp1 == NULL)
	    {
	      xlogf(DEBUG_NORMAL, FORCE, "Error with malloc of temp1 in eapttls_decode_packet().\n");
	      return -1;
	    }

          temp2 = (char *)malloc(1024);
	  if (temp2 == NULL)
	    {
	      xlogf(DEBUG_NORMAL, FORCE, "Error with malloc of temp2 in eapttls_decode_packet().\n");
	      return -1;
	    }

	  if (tls_funcs_phase == 2) 
	    {
	      eapcrypt_decrypt(in, in_size+5, temp2, &temp2_size);
	      phase2_func(temp2, temp2_size, temp1, &temp1_size);
	    } else {
	      phase2_func(out, *out_size, temp1, (int *)&temp1_size);
	    }

	  tls_funcs_phase = 2;

#if TTLS_PHASE2_DEBUG
	  printf("Going to pass this in to encrypt : \n");
	  for (i=0;i<temp1_size;i++)
	    {
	      printf("%02x ",temp1[i]);
	    }
	  printf("\n");
#endif

	  if (temp1_size>1)
	    {
	      eapcrypt_encrypt(temp1, temp1_size, temp2, &temp2_size);

	      *out_size = temp2_size;
	      memcpy(out, temp2, temp2_size);
	    }

	    if (temp2 != NULL)
	      {
		free(temp2);
		temp2 = NULL;
	      }
	    
	    if (temp1 != NULL)
	      {
		free(temp1);
		temp1 = NULL;
	      }
	    rtnVal = 0;
	}
      break;
//---------------------------------------------------------------------------
    case (0x00): 
      if (((eapcrypt_state() == 0x0003) && (phase2_func != NULL)) && (in_size>0))
	{
	  tls_funcs_phase = 2;

	  temp1 = (char *)malloc(1024);
	  if (temp1 == NULL)
	    {
	      xlogf(DEBUG_NORMAL, FORCE, "Error with malloc of temp1 in eapttls_decode_packet().\n");
	      return -1;
	    }

          temp2 = (char *)malloc(1024);
	  if (temp2 == NULL)
	    {
	      xlogf(DEBUG_NORMAL, FORCE, "Error with malloc of temp2 in eapttls_decode_packet().\n");
	      return -1;
	    }
	  //          ttls_do_phase2(temp1, &temp1_size);
	  //	  if (my_phase2 != NULL) my_phase2(temp1, &temp1_size);
	  if (tls_funcs_phase == 2) 
	    {
	      // Why in_size+3?
	      eapcrypt_decrypt(&in[1], in_size+3, temp2, &temp2_size);
	      phase2_func(temp2, temp2_size, temp1, &temp1_size);
	    } else {
	      phase2_func(out, *out_size, temp1, &temp1_size);
	    }

#if TTLS_PHASE2_DEBUG
	  printf("Going to pass this in to encrypt : \n");
	  for (i=0;i<temp1_size;i++)
	    {
	      printf("%02x ",temp1[i]);
	    }
	  printf("\n");
#endif

	  if (temp1[0] != 0x00)
	    {
	      eapcrypt_encrypt(temp1, temp1_size, temp2, &temp2_size);

	      *out_size = temp2_size;
	      memcpy(out, temp2, temp2_size);

	      if (temp2 != NULL)
		{
		  free(temp2);
		  temp2 = NULL;
		}
	    
	      if (temp1 != NULL)
		{
		  free(temp1);
		  temp1 = NULL;
		}
	    }
	  return 0;
	}
      // We got an ACK, so return some of our data.
      if (in_size > 1)
	{
	  p = in +1;

	  if (save_data_fragment(p, in_size+3) != 0)
	    {
	      xlogf(DEBUG_NORMAL, FORCE, "(TTLS) Couldn't save final data fragment!\n");
	      return -1;
	    } else {
	      xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "(TTLS) Saved final data fragment!\n");
	    }
	  
	  tptr = get_data_fragment();
	  tcnt = get_data_frag_size();

	  if (tcnt != tls_funcs_cert_size)
	    {
	      xlogf(DEBUG_NORMAL, FORCE, "(TTLS) Not enough data to form the certificate!!! %d != %d\n",tls_funcs_cert_size,tcnt);

	    }

	  rtnVal = eapcrypt_tls_parse_data((char *)get_data_fragment(), get_data_frag_size());
	}
     	rtnVal = eapcrypt_tls_return_data(out, out_size);
      	destroy_data_frags();

      break;
//---------------------------------------------------------------------------
    default:
      // we should never get here, return an error and do 
      // absolutly nothing...this packet has funky flags
#if TTLS_DEBUG
      xlogf(DEBUG_AUTHTYPES, FORCE, "(EAPTTLS) Recieved packet with incorrect flags!\n");
#endif
      rtnVal = -1;
      break;
  }

  return rtnVal;
}

int tls_funcs_build_ack(u_char *out, int *out_size)
{
  *out_size = 1;
  out = 0x00;

#if TTLS_DEBUG
  xlogf(DEBUG_AUTHTYPES, DONT_FORCE, "(EAPTTLS) Sending an ack\n");
#endif 
  return 0;
}

