/**
 * A client-side 802.1x implementation supporting EAP/TLS
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 *
 * Copyright (C) 2002 Bryan D. Payne & Nick L. Petroni Jr.
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
 * Functions to manipulate a basic configuration file
 * 
 * File: configparse.c
 *
 * Authors: bdpayne@cs.umd.edu, npetroni@cs.umd.edu
 *
 * $Id: configparse.c,v 1.15 2003/10/05 04:39:51 chessing Exp $
 * $Date: 2003/10/05 04:39:51 $
 * $Log: configparse.c,v $
 * Revision 1.15  2003/10/05 04:39:51  chessing
 * Various fixes in config parse code.
 *
 * Revision 1.14  2003/09/28 05:15:39  chessing
 * Patch to allow "default" network configuration to work again.
 *
 * Revision 1.13  2003/06/16 05:02:52  chessing
 * Code clean-ups, and documentation of functions.
 *
 * Revision 1.12  2003/06/11 21:18:18  chessing
 * Cleaned up a lot of memory leaks, added hooks for MEMWATCH to debug.
 *
 * Revision 1.11  2003/06/08 02:57:19  chessing
 * Added logging to file support, error checking after all mallocs.
 *
 * Revision 1.10  2003/03/01 06:55:39  npetroni
 * modified config parse to use lex and yacc for parsing. this may break stuff temporarily, but is good in the long run
 *
 * Revision 1.9  2003/02/11 00:25:21  galimorerpg
 * GCC 2.95.3 Fixes.
 *
 * Revision 1.8  2003/02/10 00:16:51  galimorerpg
 * Configuration parsing checks.  We now recover from some potential segfaults.
 *
 * Revision 1.7  2003/02/03 18:03:33  galimorerpg
 * ChangeLog updates.
 * Better errorchecking in configparse.c
 *
 * Revision 1.6  2003/01/14 23:52:07  chessing
 * More work on the TLS code.  It should be mostly stable now.  There is a problem if get_pass("") in eaptls_auth_challenge is called twice.
 *
 * Revision 1.5  2003/01/09 20:50:15  chessing
 * Code cleanups.
 *
 * Revision 1.4  2003/01/09 20:17:25  galimorerpg
 * Logging Updates
 *
 * Revision 1.3  2003/01/03 22:25:35  chessing
 * Turned off debugging code, (use the configure options to turn on the debugging) and added -w option to get around some problems with the Intel iANS drivers, and some wired/wireless issues with the MacOS-X wireless code.
 *
 * Revision 1.2  2003/01/02 19:29:37  chessing
 * Update to bring the xsupplicant code current with the development work.
 *
 *******************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "configparse.h"
#include "userconf.h"
#include "logging.h"


#ifndef DEBUG_CONFIGPARSE
#define DEBUG_CONFIGPARSE 0
#endif

#if MEMWATCH
#include "memwatch.h"
#endif

extern FILE *yyin;
extern int config_linenum;
extern char *config_netname;
extern const char *config_fname;

extern int yyparse(void);

// Check to see if we have config values for the network "netname".
// This could thrash the disk, and should be replaced with something more
// efficient.
int has_config(char *infile, char *netname)
{
  FILE *testfile;
  char *linein, *mysub = NULL, *linein2;

  if (infile == NULL)
    {
      xlogf(DEBUG_NORMAL, FORCE, "No file named in call for has_config()\n");
      return -1;
    }

  testfile = fopen(infile, "r");
  if (testfile == NULL)
    {
      xlogf(DEBUG_NORMAL, FORCE, "Couldn't open file %s!\n",infile);
      return -1;
    }

  linein = (char *)malloc(1024);
  if (linein == NULL) return -1;

  while (fgets(linein, 1024, testfile) != NULL)
    {
      mysub = strsep(&linein, ":");
      if ((mysub != NULL) && (linein != NULL))
	{
	  if (memcmp(netname, mysub, strlen(netname)) == 0)
	    {
	      // We have config information for this net.
	      free(mysub);
	      fclose(testfile);
	      return 1;           // We have config info.
	    }
	}

      if (mysub != NULL) free(mysub);   // Replaces the pointer for linein.
      linein = NULL;
      linein=(char *)malloc(1024);
      if (linein == NULL) return -1;
    }

  free(linein);
  linein = NULL;
  fclose(testfile);
  return 0;
}

// The name of this function is somewhat misleading.  It does parse all of the
// config values, however, the intent is to fill in the userconf structure
// for the profile (netname) we are using.
int parse_all_configs(char *infile, char *netname)
{
  if (infile == NULL)
    {
      xlogf(DEBUG_NORMAL, FORCE, "No file named in call for parse_all_configs()\n");    
    }

#if DEBUG_CONFIGPARSE
  xlogf(DEBUG_CONFIG, DONT_FORCE, "Reading configuration from file : %s\n",infile);
#endif
  
  yyin = fopen(infile, "r");
  if (yyin == NULL) { 
    xlogf(DEBUG_NORMAL, FORCE, "Could not open file %s in parse_all_configs()\n", 
	  infile); 
  }    
  
  config_netname = netname;
  if (yyparse() == -1) {
    fclose(yyin);
    return -1;
  }
  fclose(yyin);

#if DEBUG_CONFIGPARSE
  print_userconf();
#endif 

  return 0;  //If we reach here, things are good.
  
}
/*** EOF ***/
