/**
 * A client-side 802.1x implementation supporting EAP/TLS
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 *
 * Copyright (C) 2002 Bryan D. Payne & Nick L. Petroni Jr.
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
 * EAP Header for supplicant 
 * 
 * File: eap.h
 *
 * Authors: bdpayne@cs.umd.edu, npetroni@cs.umd.edu
 *
 * $Id: eap.h,v 1.6 2003/09/18 23:03:38 chessing Exp $
 * $Date: 2003/09/18 23:03:38 $
 *******************************************************************/
/***
 *** Specifics for EAP can be found in  
 *** IETF RFC 2284
 ***/

#ifndef _EAP_H
#define _EAP_H
#include "auth_methods/tls/eaptls.h"
#include "auth_methods/md5/eapmd5.h"
#include "auth_methods/ttls/eapttls.h"
#include "auth_methods/mschapv2/eapmschapv2.h"
#include "auth_methods/peap/eappeap.h"
#include "auth_methods/sim/eapsim.h"

/*********** MACRO DEFINITIONS ******************/
#define NO_EAP_AUTH -1

/* Success and Failure datagrams have form CODE ID LEN */

/* Request and Response datagrams have form CODE ID LEN TYPE */
#define EAP_TYPE_HDR          0x5       /* EAP with type 5 bytes */
/** EAP CODE FIELD - RFC 2284 Sec 2.1**/
#define EAP_REQUEST           0x1       /* Request code for EAP */
#define EAP_RESPONSE          0x2       /* Response code for EAP */
#define EAP_SUCCESS           0x3       /* Success Code */
#define EAP_FAILURE           0x4       /* Failure Code */
/** EAP Type Field - RFC 2284 Sec 3*/
#define EAP_TYPE_ID           0x1       /* Identity Type *//* REQUIRED */
#define EAP_TYPE_NOTIFY       0x2       /* Notification Type *//* REQUIRED*/
#define EAP_TYPE_NAK          0x3       /* Nak Response *//* REQUIRED */
#define EAP_TYPE_MD5          0x4       /* MD5 Challenge *//*REQUIRED*/

#define EAP_PAYLOAD_OFFSET    23        /* How many bytes in to the frame */
                                        /* before we get to the data section*/

/*********** STRUCTURE DEFINITIONS **************/

/** EAP Header with TYPE field */
struct eap_type_hdr
{
  u_char code; /* EAP Code RFC 2284 Sec. 2.1 */
  u_char id;   /* EAP identifier */
  u_short len; /* EAP data length */
  u_char type;
};

/*********** FUNCTION PROTOTYPES ****************/
/**
 * Prepares EAP package for use 
 * 
 * return: 0 if success  or -1 if fail
 */
int  
init_eap(char *, char *);

// Shutdown and clean up.
int shutdown_eap();

// Build the response frame.  Returns -1 on error.
int eap_build_responseId(u_char *out, int *out_size, int response_id);

// Build the response frame for the authentication. Returns -1 on error.
int eap_build_auth_response(u_char *in, int in_size, u_char *out, int *out_size);

/**
 * Function to handle packets and manage state machine
 * for EAP
 *
 * return -1 if fails
 *         0 if EAP request non-ID received
 *         1 if EAP request ID received
 *         3 if EAP SUCCESS received
 *         4 if EAP FAILURE received
 */
int
eap_decode_packet(u_char * /* The incoming eap packet */
		  );
  


char *return_known_eap_types();
int is_valid_eap(char *in_eap);
#endif /* _EAP_H_ */

/*** EOF ***/

