#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#if defined(LINUX)
#include <net/if_arp.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#if !defined(LINUX)
#include <sys/time.h>
#include <sys/file.h>
#include <net/bpf.h>
#endif

#include <net/if.h>

#include "logging.h"
#include "int_hooks.h"
#include "dot1x_globals.h"
#include "frame_handlers/os_frame_funcs.h"

#if MEMWATCH
#include "memwatch.h"
#endif

int int_num, sd, done = 0;
struct int_data *int_head=NULL, *int_cur=NULL, *gni = NULL;

void handle_ptr(struct int_data *indata)
{
  if (indata == NULL) return;

  if (indata->next != NULL) handle_ptr(indata->next);
  kill(indata->pid, SIGKILL);
  free(indata);
}

// Return a common seperated list of interfaces.
char *int_list()
{
  char comma[2] = ",\0";
  char *retlist = NULL;

  int_cur = int_head;
  
  retlist = (char *)malloc(1000);  // This would be a LOT of interfaces!
  if (retlist == NULL)
    {
      xlogf(DEBUG_NORMAL, DONT_FORCE, "Couldn't return a list of interfaces!\n");
      return NULL;
    }

  bzero(retlist, 1000);

  if (get_daemon()!=1)  // We only have one interface to return.
    {
      strcpy(retlist, (char *)get_dev_name());
      return retlist;
    }

  strcpy(retlist, int_cur->int_name);
  int_cur = int_cur->next;

  while (int_cur != NULL)
    {
      strcat(retlist, (char *)&comma);
      strcat(retlist, int_cur->int_name);
      int_cur = int_cur->next;
    }
  return retlist;
}

void clean_ints_and_pids()
{
  handle_ptr(int_head);
}

int add_int_to_table(char *interface)
{
  if (interface == NULL) return -1;

  if (int_head == NULL)
    {
      int_head = (struct int_data *)malloc(sizeof(struct int_data));
      if (int_head == NULL) return -1;
      int_cur = int_head;
    } else {
      int_cur = int_head;
      
      while (int_cur->next != NULL)
	{
	  int_cur = int_cur->next;
	}

      int_cur->next = (struct int_data *)malloc(sizeof(struct int_data));
      if (int_cur->next == NULL) return -1;

      int_cur = int_cur->next;
    }

  int_cur->int_name = (char *)malloc(strlen(interface)+1);
  if (int_cur->int_name == NULL) return -1;
  strncpy(int_cur->int_name, interface, strlen(interface)+1);

  int_cur->shm_id = 0;
  int_cur->pid = 0;
  int_cur->next = NULL;
  return 0;
}


int update_int_table(char *interface, int shm_id, int in_pid)
{
  int_cur = int_head;
  
  while ((int_cur != NULL) && (strncmp(int_cur->int_name, interface, strlen(int_cur->int_name)) != 0))
    {
      int_cur = int_cur->next;
    }

  if (int_cur == NULL) return -1; // We can't find the interface.

  int_cur->shm_id = shm_id;
  int_cur->pid = in_pid;
  return 0;
}

int get_interface_shmid(char *interface)
{
  int_cur = int_head;
  if ((int_cur == NULL) || (interface == NULL)) 
    {
      xlogf(DEBUG_NORMAL, FORCE, "Not enough information to return shm_id!\n");
      return -1;
    }

  while ((int_cur != NULL) && (strncmp(interface, int_cur->int_name,strlen(int_cur->int_name))!=0))
    {
      int_cur = int_cur->next;
    }

  if (int_cur == NULL)
    {
      xlogf(DEBUG_NORMAL, FORCE, "Couldn't find interface %s!\n", interface);
      return -1;
    }

  return int_cur->shm_id;
}

void init_interface_lookup()
{
#if !defined(LINUX)
  char *dev = "/dev/bpf0";
#endif

  int_num = 1;
  
  // For *BSDish systems, we don't do anything here.
#if defined(LINUX)
  sd = socket(PF_PACKET, SOCK_RAW, 0);
#else
  sd = open(dev, O_RDWR);
#endif
}

void cleanup_interface_lookup()
{
  close(sd);
}

int is_real_int(char *intname)
{
  int_cur = int_head;

  while ((int_cur != NULL) && (strncmp(int_cur->int_name, intname, strlen(int_cur->int_name))))
    {
      int_cur = int_cur->next;
    }

  if (int_cur == NULL) return 0;

  return 1;
}

int validate_int(char *intname)
{
#if !defined(LINUX)
  unsigned int devtype;
#endif
  struct ifreq ifr;

  strncpy(ifr.ifr_name, intname, sizeof(intname)+1);

#if defined(LINUX)
  if (ioctl(sd, SIOCGIFHWADDR, &ifr) < 0)
    {
      xlogf(DEBUG_NORMAL, DONT_FORCE, "Couldn't get information for interface %s!\n",intname);
    } else {
      switch (ifr.ifr_hwaddr.sa_family)
	{
	case ARPHRD_ETHER:
	case ARPHRD_IEEE80211:
	  return 1;
	}
    }
#else
  if (ioctl(sd, BIOCSETIF, (caddr_t)&ifr) < 0)
    {
      xlogf(DEBUG_NORMAL, DONT_FORCE, "Couldn't get information for interface %s!\n",intname);
    } else {
      if (ioctl(sd, BIOCGDLT, (caddr_t)&devtype) < 0)
	{
	  xlogf(DEBUG_NORMAL, DONT_FORCE, "Couldn't get information for interface %s!\n",intname);
	} else {
	  switch (devtype)
	    {
	    case DLT_EN10MB:
	    case DLT_IEEE802:
	      return 1;
	    }
	}
    }
#endif
  return 0;
}

void build_int_cache()
{
  char *int_name;

  int_name = (char *)malloc(16);  // Should never be longer.
 
  while (if_indextoname(int_num, int_name) != NULL)
    {
      int_num++;
      if (validate_int(int_name) == 1) 
	{
	  add_int_to_table(int_name);
	}
    }

  free(int_name);
  int_name = NULL;
}

char *get_next_interface()
{
  char *retVal = NULL;

  if (done == 1)
    {
      done = 0;
      return NULL;
    }

  if ((gni == NULL) && (done != 1)) 
    {
      if (int_head == NULL) return NULL;
      gni = int_head;
      done = 0;
    }
  
  retVal = gni->int_name;
  gni = gni->next;

  if (gni == NULL) done = 1;   // Flag this, so we don't start over.
  return retVal;
}
