/**
 * A client-side 802.1x implementation supporting EAP/TLS
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 *
 * Copyright (C) 2002 Bryan D. Payne & Nick L. Petroni Jr.
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
 * Hardware dependent functions for Wireless extensions
 * 
 * File: eapol-wirelessext.c
 *
 * Authors: bdpayne@cs.umd.edu, npetroni@cs.umd.edu
 *
 * $Id: os_wireless_ext.c,v 1.11 2003/10/14 04:17:08 chessing Exp $
 * $Date: 2003/10/14 04:17:08 $ 
 *******************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <linux/wireless.h>
#include <sys/ioctl.h>
#include "wireless/os_wireless_ext.h"
#include <unistd.h>
#include "logging.h"

#if MEMWATCH
#include "memwatch.h"
#endif

static const unsigned char ORINOCO_1[] = {0x00, 0x60, 0x1D};
static const unsigned char ORINOCO_2[] = {0x00, 0x02, 0x2D};
/* DKS: Hack -- on prism2 cards, must reset encryption when we roam,
   otherwise the prism2 cards attempt to encrypt their outgoing
   EAP frames, and cannot communciate with the APs. Requires
   doing something different on the orinoco cards and prism2
   cards; for now do that by recognizing the MAC address. Don't
   know what happens on non-orinoco, non-prism2 cards.
   Basically, if you turn encryption off on an orinoco card, it
   resets the card, which can cause problems. Instead, really
   want to set the orinoco card's key to '0000-0000-00' which is
   a magic cookie for "go ahead and do open system authentication". */
int
eapol_wireless_reset_encryption(char *device, u_char *mac)
{
  int rc = 0;
  if (memcmp(mac, ORINOCO_1, sizeof(ORINOCO_1)) &&
	  memcmp(mac, ORINOCO_2, sizeof(ORINOCO_2))) {
    // not an orinoco card -- reset encryption
    int ret;
    int skfd;
    struct iwreq wrq;

    memset(&wrq, 0, sizeof(struct iwreq));
    skfd = socket(AF_INET, SOCK_DGRAM, 0);
    strncpy(wrq.ifr_name, device, IFNAMSIZ);
    wrq.u.data.flags = IW_ENCODE_DISABLED;

    if ((ret = ioctl(skfd, SIOCSIWENCODE, &wrq)) < 0) {
      printf("Failed to turn off WEP: error %d\n", -ret);
      rc =  -1;
    }

    close(skfd);
  }
#ifdef EAPOL_DEBUG
  else {
    // orinoco card -- do nothing
    printf("Orinoco: ignoring ENCRYPTION RESET.\n");
  }
#endif
  printf("ENCRYPTION RESET!\n");
  return rc;
}


int
eapol_wireless_get_bssid(char *device, u_char *bssid)
{
  int skfd;            /* the socket */
  struct iwreq wrq;

  /* Get a socket */
  skfd = socket(AF_INET, SOCK_DGRAM, 0);
  strncpy(wrq.ifr_name, device, IFNAMSIZ);
  if (ioctl(skfd, SIOCGIWAP, &wrq) >= 0) {
    memcpy(bssid, wrq.u.ap_addr.sa_data, 6);
  }
  else {
    return -1;
  }
  close(skfd);
  return 0;
}

int
eapol_wireless_get_ssid(char *device, char *ssid)
{
  int skfd;            /* the socket */
  struct iwreq wrq;
  
  /* Get a socket */
  skfd = socket(AF_INET, SOCK_DGRAM, 0);  
  strncpy(wrq.ifr_name, device, IFNAMSIZ);
  wrq.u.essid.pointer = (caddr_t) ssid;
  wrq.u.essid.length=100;            // Need to specify how big the buffer is.
  wrq.u.essid.flags = 0;
  if (ioctl(skfd, SIOCGIWESSID, &wrq) < 0) {
    return -1;
  }
  close(skfd);
  return 0;
}

/* This function graciously provided by 
 Denis Belanger <denis.belanger@colubris.com>
 along with the significant crypto functionality in eapcryt.c
*/
int
eapol_wireless_set_key(char *device, u_char *key, int keylen, int index)
{
  int rc = 0;
  int skfd;
  struct iwreq wrq;

  skfd = socket(AF_INET, SOCK_DGRAM, 0);

  strncpy(wrq.ifr_name, device, IFNAMSIZ);

  wrq.u.data.flags = ((index & 0x7f) + 1) & IW_ENCODE_INDEX;
  /* DKS -- Avaya AP3s send all EAP messages in the clear, as do
   new APs when you roam. This means that cards using drivers that
   default to rejecting unencrypted packets are not able to roam,
   and are not able to respond to key updates or requests for
   reauthentication. orinoco_cs ignores this flag, this setting
   works on hostap. Not sure about wlan-ng. May introduce
   vulnerability to spoofed cleartext data packets appearing to
   come from AP. */
  wrq.u.data.flags |= IW_ENCODE_OPEN;

  wrq.u.data.length = keylen;
  wrq.u.data.pointer = (caddr_t)key;

  if ((rc = ioctl(skfd, SIOCSIWENCODE, &wrq)) < 0) {
    xlogf(DEBUG_NORMAL, FORCE, "Failed to set WEP key [%d], error %d\n",
		  (index & 0x7f) + 1, rc);
    rc =  -1;
  }
  else {
	xlogf(DEBUG_NORMAL, DONT_FORCE, "Successfully set WEP key "
		  " [%d]\n", (index & 0x7f)+1);
    if (index & 0x80) {
      /* This is an unicast key: use it for transmission */
      strncpy(wrq.ifr_name, device, IFNAMSIZ);

      wrq.u.data.flags = ((index & 0x7f) + 1) & IW_ENCODE_INDEX;
      /* DKS - see above */
      wrq.u.data.flags |= IW_ENCODE_OPEN;

      wrq.u.data.length = 0;
      wrq.u.data.pointer = (caddr_t)NULL;

      if (ioctl(skfd, SIOCSIWENCODE, &wrq) < 0) {
	xlogf(DEBUG_NORMAL, DONT_FORCE, "Failed to set the WEP transmit key ID [%d]\n", (index & 0x7f)+1);
	rc =  -1;
      }
      else
	xlogf(DEBUG_NORMAL, DONT_FORCE, "Successfully set the WEP transmit key "
	      " [%d]\n", (index & 0x7f)+1);
    }
  }

  close(skfd);
  return rc;
}

int eapol_wireless_init()
{
  return 0;
}

int eapol_wireless_cleanup()
{
  return 0;
}

/*** EOF ***/





