/**
 * A client-side 802.1x implementation supporting EAP/TLS
 *
 * This code is released under both the GPL version 2 and BSD licenses.
 * Either license may be used.  The respective licenses are found below.
 *
 * Copyright (C) 2002 Bryan D. Payne & Nick L. Petroni Jr.
 * All Rights Reserved
 *
 * --- GPL Version 2 License ---
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * --- BSD License ---
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  - All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by the University of
 *       Maryland at College Park and its contributors.
 *  - Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*******************************************************************
 * The driver function for a Linux application layer EAPOL 
 * implementation
 * File: userconf.c
 *
 * Authors: Chris.Hessing@utah.edu, Terry.Simons@utah.edu
 *
 * $Id: userconf.c,v 1.28 2003/07/05 06:22:28 chessing Exp $
 * $Date: 2003/07/05 06:22:28 $
 * $Log: userconf.c,v $
 * Revision 1.28  2003/07/05 06:22:28  chessing
 * Fixes to int_hooks, and userconf stuff.
 *
 * Revision 1.27  2003/07/05 05:05:22  chessing
 * Some changes to hardware inspection code, to fix some strange bugs.
 *
 * Revision 1.26  2003/07/03 04:53:06  chessing
 * Push updates for GUIs, a few strange bug fixes.
 *
 * Revision 1.25  2003/07/02 03:40:45  chessing
 * We now check the essid each time through the authentication loop.  This should allow seamless roaming between networks with different essids.
 *
 * Revision 1.24  2003/07/02 02:33:54  chessing
 * Moved password prompt before we send the initial responseId, and removed all password prompts from various EAP methods.
 *
 * Revision 1.23  2003/07/01 05:02:43  chessing
 * Code cleanups, added config option that will seperate phase 1, and phase 2 username.  Allows use of "Anonymous" in phase 1 response ID.
 *
 * Revision 1.22  2003/06/29 21:31:23  chessing
 * Various memory leak cleanups.
 *
 * Revision 1.21  2003/06/29 05:38:48  chessing
 * Various memory leak cleanups, first hack at PEAP version 0, works with Radiator!
 *
 * Revision 1.20  2003/06/22 00:44:54  chessing
 * TLS related code merged in to a single file.  Should allow for easier implementation of future TLS based authentication methods.
 *
 * Revision 1.19  2003/06/19 02:32:00  chessing
 * Additions to better support daemon mode.  Updates to the gui-commands files.
 *
 * Revision 1.18  2003/06/16 03:55:44  chessing
 * GUI interface now allows setting of values in supplicant, in flight.
 *
 * Revision 1.17  2003/06/15 23:16:19  chessing
 * Bug fixes, and GUI Tools additions.
 *
 * Revision 1.16  2003/06/15 20:30:05  chessing
 * Children now die when parent is killed, in daemonized mode.
 *
 * Revision 1.15  2003/06/14 20:39:07  chessing
 * Added the following :
 * 	New daemon mode.  Will fork for each real interface, and start a state machine.
 * 	New GUI interface. Still needs work, but has the basics to do a few things.  Common GUI tools source can be found in the gui_tools directory.
 * 	Uses shared memory if available.  If shared memory isn't available, daemon mode is disabled.
 *
 * Revision 1.14  2003/06/12 23:35:33  chessing
 * Fix to os_wireless_ext for linux to allow Prism cards to work right.
 *
 * Revision 1.13  2003/06/12 03:33:03  chessing
 * Fixes bug where default settings would bleed through when a different network name was found, and the net network name doesn't have a value defined.
 *
 * Revision 1.12  2003/06/11 21:18:20  chessing
 * Cleaned up a lot of memory leaks, added hooks for MEMWATCH to debug.
 *
 * Revision 1.11  2003/06/08 02:57:19  chessing
 * Added logging to file support, error checking after all mallocs.
 *
 * Revision 1.10  2003/05/20 01:44:47  chessing
 * Fixed problems with the -u option.  Added patch for rekeying with Agere based wireless cards.  Cleaned up userconf.c.  And, added first hack at TTLS-CHAP. (Still doesn't work. ;)
 *
 * Revision 1.9  2003/03/28 22:54:02  chessing
 * Patch to fix endianess issues between PPC and x86, with TTLS-PAP.
 *
 * Revision 1.8  2003/03/20 20:14:40  chessing
 * Many patches from Xerox PARC.  They include fixes to show-stoppers, better roaming support, CRLs and chain verification, patch to allow spaces in ESSIDs, some state machine timing issues, and prism2 card fixes.
 *
 * Revision 1.7  2003/03/13 22:31:11  chessing
 * Another attempt at the wired/wireless fix.
 *
 * Revision 1.6  2003/03/04 06:10:08  npetroni
 * changes to allow a program to be run after successful authentication
 *
 * Revision 1.5  2003/03/01 06:55:39  npetroni
 * modified config parse to use lex and yacc for parsing. this may break stuff temporarily, but is good in the long run
 *
 * Revision 1.4  2003/01/14 23:52:07  chessing
 * More work on the TLS code.  It should be mostly stable now.  There is a problem if get_pass("") in eaptls_auth_challenge is called twice.
 *
 * Revision 1.3  2003/01/09 21:01:58  galimorerpg
 * Logging Updates and Build Fixes.
 *
 * Revision 1.2  2003/01/03 22:25:35  chessing
 * Turned off debugging code, (use the configure options to turn on the debugging) and added -w option to get around some problems with the Intel iANS drivers, and some wired/wireless issues with the MacOS-X wireless code.
 *
 * Revision 1.1  2003/01/02 19:29:37  chessing
 * Update to bring the xsupplicant code current with the development work.
 *
 *
 *******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifndef NO_DAEMON
#include <sys/shm.h>
#endif

#include "userconf.h"
#include "logging.h"
#include "dot1x_globals.h"
#include "1xdaemon.h"

#if MEMWATCH
#include "memwatch.h"
#endif

#ifndef DEBUG_USERCONF
#define DEBUG_USERCONF 0
#endif

struct userconf *user_settings=NULL;
int cur_mem_chunk = 0;              // The shared memory id we are currently
                                    // using.  0 means none.

// Set up a message for the controlling daemon to send.
void set_push_msg(uint8_t msgId, char *msgtosend)
{
  // First, clear out anything that is already there.
  bzero(user_settings->push_msg, 255);  

  user_settings->push_msg[0] = msgId;  // The message to send.
  
  strncpy(&user_settings->push_msg[1], msgtosend, 255);
}

// This function returns 1 when a username and password are present.
// This allows us to avoid starting an authentication attempt in daemon mode
// that would ultimately cause a process to "freeze" waiting for a password.
int auth_ready()
{
  char *username, *password;

  username = get_username();
  password = get_password();

  if ((username == NULL) || (password == NULL)) return 0;
  xlogf(DEBUG_NORMAL, DONT_FORCE, "Username = %s   Pasword = %s\n",username, password);

  free(username);
  free(password);
  return 1;
}

void user_settings_mem_attach(int new_mem_id)
{
#ifndef NO_DAEMON
  // If we aren't a daemon, bail.
  if (get_daemon()!=1) return;

  if (cur_mem_chunk != 0)
    {
      if (cur_mem_chunk != new_mem_id)
	{
	  if (shmdt(user_settings)!=0)
	    {
	      xlogf(DEBUG_NORMAL, DONT_FORCE, "Couldn't detatch shared memory.  Things may act funny.\n");
	    }
	}
    }

  if (cur_mem_chunk != new_mem_id)
    {
      user_settings = (struct userconf *)shmat(new_mem_id, 0, SHM_W | SHM_R);
    }
#endif
}

void clear_user_settings()
{
  bzero(user_settings, sizeof(struct userconf));
  user_settings->chunk_size = 1398;
}

void initalize_user_conf(int new_mem_id)
{
  if (get_daemon()==1)
    {
      user_settings_mem_attach(new_mem_id);
    } else {
      user_settings = (struct userconf *)malloc(sizeof(struct userconf));
    }

  if (user_settings == NULL)
    {
      xlogf(DEBUG_NORMAL, FORCE, "Error on malloc in initalize_user_conf().\n");
      return;
    }

  clear_user_settings();

#if DEBUG_USERCONF
  xlogf(DEBUG_CONFIG, DONT_FORCE, "initalize_user_conf() finished!\n");
  if (user_settings == NULL) xlogf(DEBUG_CONFIG, FORCE, "user_settings returned NULL!!!!!!\n");
#endif
}

int is_ready()
{
  if ((user_settings->username != NULL) && (user_settings->password != NULL))
    {
      return 1;
    }
  return 0;
}


void clean_user_conf()
{
  clear_user_settings();

  free(user_settings);
#if DEBUG_USERCONF
  xlogf(DEBUG_CONFIG, DONT_FORCE, "clean_user_conf() called!\n");
#endif
}

void str_set(char *to_set, char *value, char *to_persist, int persist, int maxlen)
{
  if (!((to_persist != NULL) && (*to_persist == 1)))
    {
      if (value != NULL)
	{
	  strncpy(to_set, value, maxlen);
	  if (to_persist != NULL) 
	    {
	      *to_persist = persist;
	    }
	}
    }
}

char *str_get(char *getval)
{
  char *return_val=NULL;

  if ((getval == NULL) || (getval[0]=='\0')) return NULL;

  return_val = (char *)malloc(strlen(getval)+1);
  if (return_val == NULL)
    {
      xlogf(DEBUG_NORMAL, FORCE, "Error with malloc in str_get().\n");
      return NULL;
    }
  strncpy(return_val, getval, strlen(getval)+1);
  
  return return_val;
}

char *get_push_msg()
{
  return str_get((char *)&user_settings->push_msg);
}

// Reset the push_msg to all 0s.
void clear_push_msg()  
{
  bzero(&user_settings->push_msg, 255);
}

int get_statemachine_state()
{
  return user_settings->current_state;
}

void set_statemachine_state(int new_state)
{
  user_settings->current_state = new_state;
}

char *get_username()
{
  return str_get(user_settings->username);
}

void set_username(char *in_username, char persist)
{
  str_set((char *)&user_settings->username, in_username, &user_settings->user_persist, persist, 255);
}

char *get_password()
{
  return str_get(user_settings->password);
}

void set_password(char *in_password)
{
  str_set((char *)&user_settings->password, in_password, NULL, 0, 255);
}

char *get_root_cert()
{
  return str_get(user_settings->root_cert);
}

void set_root_cert(char *in_root_cert)
{
  str_set((char *)&user_settings->root_cert, in_root_cert, NULL, 0, 512);
}

char *get_client_cert()
{
  return str_get(user_settings->client_cert);
}

void set_client_cert(char *in_client_cert)
{
  str_set((char *)&user_settings->client_cert, in_client_cert, NULL, 0, 512);
}

char *get_root_dir()
{
  return str_get(user_settings->root_dir);
}

void set_root_dir(char *in_root_dir)
{
  str_set((char *)&user_settings->root_dir, in_root_dir, NULL, 0, 512);
}

char *get_crl_dir()
{
  return str_get(user_settings->crl_dir);
}

void set_crl_dir(char *in_crl_dir)
{
  str_set((char *)&user_settings->crl_dir, in_crl_dir, NULL, 0, 512);
}

char *get_key_file()
{
  return str_get(user_settings->key_file);
}

void set_key_file(char *in_key_file)
{
  str_set((char *)&user_settings->key_file, in_key_file, NULL, 0, 512);
}

char *get_auth()
{
  return str_get(user_settings->auth);
}

void set_auth(char *in_auth)
{
  str_set((char *)&user_settings->auth, in_auth, NULL, 0, 100);
}

char *get_preferred_auth()
{
  return str_get(user_settings->preferred_auth);
}

// We need to read the string in and set it to upper case.
void set_preferred_auth(char *in_preferred_auth)
{
  int i=0;

  if (in_preferred_auth != NULL)
    {
      // Make sure the value we are passed is in all CAPS!
      for (i=0; i<=strlen(in_preferred_auth); i++) 
	{
	  in_preferred_auth[i] = toupper(in_preferred_auth[i]);
	}
      str_set((char *)&user_settings->preferred_auth, in_preferred_auth, NULL, 0, 25);
    }
}

char *get_client_type()
{
  return str_get(user_settings->client_type);
}

void set_client_type(char *in_client_type)
{
  int i=0;
  char *tempstr;

  if (in_client_type != NULL) 
    {
      tempstr = (char *)malloc(strlen(in_client_type)+1);
      if (tempstr == NULL)
	{
	  xlogf(DEBUG_NORMAL, FORCE, "Error with malloc in set_client_type().\n");
	  return;
	}
      strncpy(tempstr, in_client_type, strlen(in_client_type)+1);
      // Make sure the value we are passed is in all CAPS!
      for (i=0; i<=strlen(tempstr); i++)
	{
	  tempstr[i] = toupper(tempstr[i]);
	}
      str_set((char *)&user_settings->client_type, tempstr, NULL, 0, 10);
      free(tempstr);
      tempstr = NULL;
    }
}

int get_chunk_size()
{
  return user_settings->chunk_size;
}

void set_chunk_size(int new_size)
{
  if (new_size<=0)
    {
      xlogf(DEBUG_CONFIG, FORCE, "Chunk size must be greater than 0!\n");
      return;
    }

  user_settings->chunk_size = new_size;
}

char *get_random_file()
{
  return str_get(user_settings->random_file);
}

void set_random_file(char *new_random_file)
{
  str_set((char *)&user_settings->random_file, new_random_file, NULL, 0, 512);
}

char *get_first_auth()
{
  return str_get(user_settings->first_auth);
}

void set_first_auth(char *new_first_auth)
{
  str_set((char *)&user_settings->first_auth, new_first_auth, NULL, 0, 512);
}

char *get_after_auth()
{
  return str_get(user_settings->after_auth);
}

void set_after_auth(char *new_after_auth)
{
  str_set((char *)&user_settings->after_auth, new_after_auth, NULL, 0, 512);
}

char *get_phase2auth()
{
  return str_get(user_settings->phase2auth);
}


void set_phase2auth(char *new_phase2auth)
{
  str_set((char *)&user_settings->phase2auth, new_phase2auth, NULL, 0, 50);
}

char *get_phase2id()
{
  char *phase2_id_ret;

  phase2_id_ret = str_get(user_settings->phase2id);

  if (phase2_id_ret == NULL) return str_get(user_settings->username);

  // Otherwise, go ahead and return the "correct" value.
  return phase2_id_ret;
}

void set_phase2id(char *new_phase2id)
{
  str_set((char *)&user_settings->phase2id, new_phase2id, NULL, 0, 255);
}


void print_userconf()
{
  printf("--Printing user configuration--\n");
  printf("username: %s\n", get_username());
  printf("root cert: %s\n", get_root_cert());
  printf("client cert: %s\n", get_client_cert());
  printf("key file: %s\n", get_key_file());
  printf("auth type: %s\n", get_auth());
  printf("preferred auth: %s\n", get_preferred_auth());
  printf("client type: %s\n", get_client_type());
  printf("chunk size: %d\n", get_chunk_size());
  printf("random file: %s\n", get_random_file());
  printf("first auth: %s\n", get_first_auth());
  printf("after auth: %s\n", get_after_auth());
  printf("phase2 auth: %s\n", get_phase2auth());
  printf("phase2 id: %s\n", get_phase2id());
  printf("-------------------------------\n");
}

// Pass a value in, if set=1, then set the value, and return it, otherwise
// just return it.  (If it is a get, pass NULL for newval.)
char *get_set_userconf_value(char *inval, char *newval)
{
  char *pwdreturn;
  char *numconvert;
  int chunk_size;

  if (strcmp(inval, "username") == 0)
    {
      if (newval != NULL)  set_username(newval, 0);
      return get_username();
    }
  if (strcmp(inval, "password") == 0)
    {
      if (newval != NULL) set_password(newval);
      
      pwdreturn = (char *)malloc(512);
      if (pwdreturn == NULL) return NULL;

      strcpy(pwdreturn, "Password is not returned!");
      return pwdreturn;
    }
  if (strcmp(inval, "root") == 0)
    {
      if (newval != NULL)  set_root_cert(newval);
      return get_root_cert();
    }
  if (strcmp(inval, "cert") == 0)
    {
      if (newval != NULL)  set_client_cert(newval);
      return get_client_cert();
    }
  if (strcmp(inval, "key") == 0)
    {
      if (newval != NULL)  set_key_file(newval);
      return get_key_file();
    }
  if (strcmp(inval, "auth") == 0)
    {
      if (newval != NULL) set_auth(newval);
      return get_auth();
    }
  if (strcmp(inval, "pref") == 0)
    {
      if (newval != NULL) set_preferred_auth(newval);
      return get_preferred_auth();
    }
  if (strcmp(inval, "type") == 0)
    {
      if (newval != NULL) set_client_type(newval);
      return get_client_type();
    }
  if (strcmp(inval, "chunk_size") == 0)
    {
      if (newval != NULL) set_chunk_size(atoi(newval));
      chunk_size = get_chunk_size();
      numconvert = (char *)malloc(80);
      if (numconvert == NULL) return NULL;
      sprintf(numconvert, "%d", chunk_size);
      return numconvert;
    }
  if (strcmp(inval, "random_file") == 0)
    {
      if (newval != NULL) set_random_file(newval);
      return get_random_file();
    }
  if (strcmp(inval, "first_auth") == 0)
    {
      if (newval != NULL) set_first_auth(newval);
      return get_first_auth();
    }
  if (strcmp(inval, "after_auth") == 0)
    {
      if (newval != NULL) set_after_auth(newval);
      return get_after_auth();
    }
  if (strcmp(inval, "phase2auth") == 0)
    {
      if (newval != NULL) set_phase2auth(newval);
      return get_phase2auth();
    }
  if (strcmp(inval, "phase2id") == 0)
    {
      if (newval != NULL) set_phase2id(newval);
      return get_phase2id();
    }
  return NULL;
}

char *get_userconf_value(char *instr)
{
  return get_set_userconf_value(instr, NULL);
}

// We will be handed in a string of the format <tag>=<value>.
char *set_userconf_value(char *instr)
{
  char *tempstr, *setting, *value;

  tempstr = (char *)malloc(strlen(instr)+1);
  if (tempstr == NULL) return NULL;

  strcpy(tempstr, instr);
  setting = strtok(tempstr, "=");
  value = strtok(NULL, "=");

  lower_str(setting);  // Make it easier to match.

  return get_set_userconf_value(setting, value);
}
